# coding=utf-8
# ======================================
# File:     test_eastmoney.py
# Author:   Jackie PENG
# Contact:  jackie.pengzhao@gmail.com
# Created:  2023-11-01
# Desc:
#   Unittest for all eastmoney data
# acquiring functions and apis.
# ======================================

import unittest

import pandas as pd
import numpy as np

from qteasy.emfuncs import (
    stock_daily,
    stock_weekly,
    stock_1min,
    _stock_bars,
    real_time_klines,
    real_time_quote,
)
from qteasy.utilfuncs import is_market_trade_day


class TestEastmoney(unittest.TestCase):
    """ Test eastmoney data acquiring functions and apis """

    def SetUp(self):
        pass

    def test_get_k_history(self):
        """ Test _get_k_history function """
        code = '000651'
        date = (pd.to_datetime('today') - pd.Timedelta(days=3)).strftime('%Y%m%d')  # 最多只能获取过去10个交易日的5分钟K线
        end = pd.to_datetime('today').strftime('%Y%m%d')
        res = _stock_bars(qt_code=code, start=date, end=end, freq='5min')
        print(res)

        self.assertIsInstance(res, pd.DataFrame)
        if is_market_trade_day(end):
            self.assertFalse(res.empty, f'empty data downloaded for {code} on {end}, which is a trade day!')
        else:
            self.assertTrue(res.empty)
        code = '000651'
        res = stock_daily(qt_code=code, start='20231102', end='20240301')
        print(res)
        self.assertIsInstance(res, pd.DataFrame)
        self.assertFalse(res.empty)
        self.assertTrue(np.all(res.close == res.pre_close + res.change))
        self.assertTrue(np.all(np.abs(res.pct_chg - (res.change / res.pre_close) * 100) < 0.01))
        code = '000651.SZ'
        res = stock_weekly(qt_code=code, start='20231102', end='20240301')
        print(res)
        self.assertIsInstance(res, pd.DataFrame)
        self.assertFalse(res.empty)
        code = '000001.SH'
        res = stock_1min(qt_code=code, start='20231102', end='20240301')
        print(res)
        self.assertIsInstance(res, pd.DataFrame)
        self.assertFalse(res.empty)

    def test_real_time_kline_price(self):
        """ Test real_time_klines function """

        code = '000016.SZ'
        date = '20241227'
        res = real_time_klines(qt_code=code, freq='D', date=date)
        print(f'data acquied for code \'000016.SZ\': {res}')
        self.assertIsInstance(res, pd.DataFrame)
        from qteasy.utilfuncs import is_market_trade_day
        if is_market_trade_day(date):
            self.assertFalse(res.empty, f'empty data downloaded for {code} on {date}, which is a trade day!')
            self.assertEqual(len(res), 2)
            self.assertEqual(res.columns.to_list(),
                             ['symbol', 'name', 'pre_close', 'open', 'close', 'high', 'low', 'vol', 'amount'])
            self.assertEqual(res.index.name, 'trade_date')
            self.assertTrue(all(item in code for item in res.symbol))
            # some items may not have real time price at the moment
            # self.assertTrue(all(item in res.symbol.to_list() for item in code))
        else:
            print(f'not a trade day, no real time k-line data acquired!')
            self.assertTrue(res.empty)

        code = '000025.SZ'
        date = '20241231'
        res = real_time_klines(qt_code=code, date=date, freq='D')
        print(res)
        self.assertIsInstance(res, pd.DataFrame)

        code = '000016.SZ'
        date = (pd.to_datetime('today') - pd.Timedelta(days=3)).strftime('%Y%m%d')  # 最多只能获取过去10个交易日的5分钟K线
        res = real_time_klines(qt_code=code, date=date, freq='5min')
        print(res)
        self.assertIsInstance(res, pd.DataFrame)
        if is_market_trade_day(date):
            self.assertFalse(res.empty)
            self.assertEqual(res.columns.to_list(), ['symbol', 'name', 'pre_close', 'open', 'close',
                                                     'high', 'low', 'vol', 'amount'])
            self.assertEqual(res.index.name, 'trade_time')
            self.assertTrue(all(item in code for item in res.symbol))
            # some items may not have real time price at the moment
            # self.assertTrue(all(item in res.symbol.to_list() for item in code))
        else:
            print(f'not a trade day, no real time k-line data acquired!')
            self.assertTrue(res.empty)

        print('test acquiring Index prices')
        codes = '399001.SZ'
        date = (pd.to_datetime('today')).strftime('%Y%m%d')  # 最多只能今天交易日的1分钟K线
        res = real_time_klines(qt_code=codes, date=date, freq='1min')
        print(res)
        self.assertIsInstance(res, pd.DataFrame)
        if is_market_trade_day(date):
            self.assertFalse(res.empty)
            self.assertEqual(res.columns.to_list(), ['symbol', 'name', 'pre_close', 'open', 'close',
                                                     'high', 'low', 'vol', 'amount'])
            self.assertEqual(res.index.name, 'trade_time')
            self.assertTrue(all(item in codes for item in res.symbol))
            # some items may not have real time price at the moment
            # self.assertTrue(all(item in res.symbol.to_list() for item in code))
        else:
            print(f'not a trade day, no real time k-line data acquired!')
            self.assertTrue(res.empty)

        print('test acquiring ETF price data')
        codes = '510300.SH'
        date = (pd.to_datetime('today') - pd.Timedelta(days=3)).strftime('%Y%m%d')  # 最多只能获取过去10个交易日的60分钟K线
        res = real_time_klines(qt_code=codes, date=date, freq='h')
        print(res)
        self.assertIsInstance(res, pd.DataFrame)
        if is_market_trade_day(date):
            self.assertFalse(res.empty)
            self.assertEqual(res.columns.to_list(), ['symbol', 'name', 'pre_close', 'open', 'close',
                                                     'high', 'low', 'vol', 'amount'])
            self.assertEqual(res.index.name, 'trade_time')
            self.assertTrue(all(item in codes for item in res.symbol))
            # some items may not have real time price at the moment
            # self.assertTrue(all(item in res.symbol.to_list() for item in code))
        else:
            print(f'not a trade day, no real time k-line data acquired!')
            self.assertTrue(res.empty)

    def test_real_time_quote(self):
        """ testing function get real time quote"""
        code = '399001.SZ'
        res = real_time_quote(qt_code=code)
        print(res)
        self.assertIsInstance(res, pd.DataFrame)
        if is_market_trade_day((pd.to_datetime('today')).strftime('%Y%m%d')):
            self.assertFalse(res.empty, f'empty data downloaded for {code}, which is a trade day!')
        else:
            self.assertTrue(res.empty)


if __name__ == '__main__':
    unittest.main()



# coding=utf-8
# ======================================
# File:     data_channels.py
# Author:   Jackie PENG
# Contact:  jackie.pengzhao@gmail.com
# Created:  2024-10-09
# Desc:
#   Definition of interface to
# acquire or download historical data
# that can be stored in the datasource
# from different channels such as
# tushare, yahoo finance, akshare, etc.
# ======================================
import numpy as np
import pandas as pd
import time
import logging

from typing import Generator, Union, Any
from functools import lru_cache

from concurrent.futures import (
    ThreadPoolExecutor,
    as_completed,
)

from .utilfuncs import (
    str_to_list,
    list_truncate,
    list_to_str_format,
    get_current_timezone_datetime,
)

"""
这个模块提供一个统一数据下载api：
data_channels.download_data(
    channel=channel,
    table=table_name,
    shares=shares,
    start_date=start_date,
    end_date=end_date,
    columns=columns,
    kwargs=kwargs,
)
其中table_name为数据表名，shares为股票代码，start_date和end_date为开始和
结束日期，columns为需要的列名，kwargs为其他参数。
这个api会根据channel的不同，找到不同数据源的下载api，然后调用这个api下载数据，
下载的数据将会被去重以后保存到一个DataFrame中，并返回。

数据的column整理不在这个模块中进行，而是在database模块中进行。这里只进行row方
向的去重处理

针对不同的数据源，本模块定义不同的数据表和下载api的对照表以及参数表，根据
这些表，将采用优化的方式下载数据。

优化方式包括：
1，数据量拆分，对于数据量较大的数据，可以将数据拆分成多组小数据，分批下载，拆分的方式依据对照表。
2，并行下载，拆分后的数据可以并行下载，加快下载速度。
3，数据缓存组装，对于大量数据，进行优化组装，提高速度

"""


# =====================
# data_channel模块主要API的解析函数，根据channel解析正确的API以执行
# =====================
def _fetch_table_data_from_tushare(table, **kwargs):
    """使用kwargs参数，从tushare获取一次金融数据

    Parameters
    ----------
    table: str,
        数据表名，必须是API_MAP中定义的数据表
    **kwargs:
        用于下载金融数据的函数参数

    Returns
    -------
    pd.DataFrame:
        下载后的数据
    """

    from .tsfuncs import acquire_data
    dnld_data = acquire_data(TUSHARE_API_MAP[table][API_MAP_COLUMNS.index('api')], **kwargs)

    return dnld_data


def _fetch_table_data_from_akshare(table, **kwargs):
    """ 使用kwargs参数，从akshare获取一次金融数据

    Parameters
    ----------
    table: str,
        数据表名，必须是API_MAP中定义的数据表
    **kwargs:
        用于下载金融数据的函数参数

    Returns
    -------
    pd.DataFrame:
        下载后的数据
    """
    from .akfuncs import acquire_data
    dnld_data = acquire_data(AKSHARE_API_MAP[table][API_MAP_COLUMNS.index('api')], **kwargs)

    return dnld_data


def _fetch_table_data_from_eastmoney(table, **kwargs):
    """ 使用kwargs参数，从东方财富网获取一次金融数据

    Parameters
    ----------
    table: str,
        数据表名，必须是API_MAP中定义的数据表
    **kwargs:
        用于下载金融数据的函数参数

    Returns
    -------
    pd.DataFrame:
        下载后的数据
    """
    from .emfuncs import acquire_data
    dnld_data = acquire_data(EASTMONEY_API_MAP[table][API_MAP_COLUMNS.index('api')], **kwargs)

    return dnld_data


def _fetch_table_data_from_sina(table, **kwargs):
    """ 使用kwargs参数，从新浪财经获取一次金融数据

    Parameters
    ----------
    table: str,
        数据表名，必须是API_MAP中定义的数据表
    **kwargs:
        用于下载金融数据的函数参数

    Returns
    -------
    pd.DataFrame:
        下载后的数据
    """
    from .sinafuncs import acquire_data
    dnld_data = acquire_data(SINA_API_MAP[table][API_MAP_COLUMNS.index('api')], **kwargs)

    return dnld_data


def _fetch_realtime_kline_from_tushare(qt_code, date, freq):
    """ 从tushare获取实时K线数据"""
    from .tsfuncs import acquire_data
    api_name = TUSHARE_REALTIME_API_MAP['realtime_bars'][API_MAP_COLUMNS.index('api')]
    dnld_data = acquire_data(api_name, **{'ts_code': qt_code, 'freq': freq})

    return dnld_data


def _fetch_realtime_kline_from_akshare(qt_code, date, freq):
    """ 从akshare获取实时K线数据"""
    from .akfuncs import acquire_data
    api_name = AKSHARE_REALTIME_API_MAP['realtime_bars'][API_MAP_COLUMNS.index('api')]
    dnld_data = acquire_data(api_name, **{'qt_code': qt_code, 'date': date, 'freq': freq})

    return dnld_data


def _fetch_realtime_kline_from_eastmoney(qt_code, date, freq):
    """ 从eastmoney获取实时K线数据"""
    from .emfuncs import acquire_data
    api_name = EASTMONEY_REALTIME_API_MAP['realtime_bars'][API_MAP_COLUMNS.index('api')]
    dnld_data = acquire_data(api_name, **{'qt_code': qt_code, 'date': date, 'freq': freq})

    return dnld_data


def _fetch_realtime_kline_from_sina(qt_code, date, freq):
    """ 从新浪财经获取实时K线数据"""
    from .sinafuncs import acquire_data
    api_name = SINA_REALTIME_API_MAP['realtime_bars'][API_MAP_COLUMNS.index('api')]
    dnld_data = acquire_data(api_name, **{'qt_code': qt_code, 'date': date, 'freq': freq})

    return dnld_data


def _get_fetch_table_func(channel: str):
    """ 获取数据下载函数

    Parameters
    ----------
    channel: str,
        数据获取渠道，指定本地文件、金融数据API，或直接给出local_df，支持以下选项：
        - 'tushare'     : 从Tushare API获取金融数据，请自行申请相应权限和积分
        - 'akshare'     : 从AKshare API获取金融数据
        - 'eastmoney'   : 从东方财富网获取金融数据

    Returns
    -------
    function:
        数据下载函数
    """
    if channel == 'tushare':
        return _fetch_table_data_from_tushare
    elif channel == 'akshare':
        return _fetch_table_data_from_akshare
    elif channel in ['emoney', 'eastmoney']:
        return _fetch_table_data_from_eastmoney
    elif channel == 'sina':
        return _fetch_table_data_from_sina
    else:
        raise NotImplementedError(f'channel {channel} is not supported')


def _get_realtime_kline_func(channel: str):
    """ 获取实时K线下载函数"""
    if channel == 'tushare':
        return _fetch_realtime_kline_from_tushare
    elif channel == 'akshare':
        return _fetch_realtime_kline_from_akshare
    elif channel in ['emoney', 'eastmoney']:
        return _fetch_realtime_kline_from_eastmoney
    elif channel == 'sina':
        return _fetch_realtime_kline_from_sina
    else:
        raise NotImplementedError(f'channel {channel} is not supported')


# =====================
# data_channel模块的主要API，分别用于从不同的渠道获取数据表数据以及实时价格数据（实时数据仅包含实时价格数据，且格式统一）
# =====================
def parse_data_fetch_args(table, channel, symbols, start_date, end_date, list_arg_filter, reversed_par_seq) -> dict:
    """ 解析数据获取API的参数，生成下载数据的参数序列

    本函数为data_channel的主API进行参数解析，用户在获取数据时，一般仅会指定需要获取的数据
    类型、频率、开始和结束日期，以及可能的筛选参数，如股票代码、交易所等。Data_channel会将
    这些参数解析成一组参数序列，用于调用数据获取API。

    Parameters
    ----------
    table: str,
        数据表名，必须是database中定义的数据表
    channel: str,
        数据获取渠道，指定本地文件、金融数据API，或直接给出local_df，支持以下选项：
        - 'tushare'     : 从Tushare API获取金融数据，请自行申请相应权限和积分
        - 'akshare'     : 从AKshare API获取金融数据
        - 'eastmoney'   : 从东方财富网获取金融数据
    start_date: str, optional
        数据下载的开始日期
    end_date: str, optional
        数据下载的结束日期
    list_arg_filter: str or list of str, optional
        用于下载数据时的筛选参数，某些数据表以列表的形式给出可筛选参数，如stock_basic表，它有一个可筛选
        参数"exchange"，选项包含 'SSE', 'SZSE', 'BSE'，可以通过此参数限定下载数据的范围。
        如果filter_arg为None，则下载所有数据。
        例如，下载stock_basic表数据时，下载以下输入均为合法输入：
        - 'SZSE'
            仅下载深圳交易所的股票数据
        - ['SSE', 'SZSE']
        - 'SSE, SZSE'
            上面两种写法等效，下载上海和深圳交易所的股票数据
    symbols: str or list of str, optional
        用于下载数据的股票代码，如果给出了symbols，只有这些股
        票代码的数据会被下载
    reversed_par_seq: bool, default False
        是否将参数序列反转，如果为True，则会将参数序列反转，用于下载数据时的优化

    Returns
    -------
    dict:
        用于下载数据的参数序列
    """

    if channel == 'tushare':
        API_MAP = TUSHARE_API_MAP
    elif channel == 'akshare':
        API_MAP = AKSHARE_API_MAP
    elif channel == 'eastmoney':
        API_MAP = EASTMONEY_API_MAP
    else:
        raise NotImplementedError(f'channel {channel} is not supported')

    if table not in API_MAP:
        return {}

    # get all tables in the API mapping
    arg_name = API_MAP[table][1]
    arg_type = API_MAP[table][2]
    arg_range = API_MAP[table][3]
    allowed_code_suffix = API_MAP[table][4]
    additional_start_end = API_MAP[table][5]
    start_end_chunk_size = API_MAP[table][6]

    if isinstance(symbols, list):
        symbols = list_to_str_format(symbols)

    # parse the filling args and pick the first filling arg value from the range

    if arg_name == 'none':
        arg_name = None

    if arg_type == 'list':
        arg_values = _parse_list_args(arg_range, list_arg_filter, reversed_par_seq)
    elif arg_type == 'datetime':
        from .datatables import TABLE_MASTERS
        freq = TABLE_MASTERS[table][4]
        arg_values = _parse_datetime_args(arg_range, start_date, end_date, freq, reversed_par_seq)
    elif arg_type == 'trade_date':
        from .datatables import TABLE_MASTERS
        freq = TABLE_MASTERS[table][4]
        arg_values = _parse_trade_date_args(arg_range, start_date, end_date, freq, 'SSE', reversed_par_seq)
    elif arg_type == 'hk_trade_date':
        raise NotImplementedError('hk_trade_date is not implemented')
    elif arg_type == 'us_trade_date':
        raise NotImplementedError('us_trade_date is not implemented')
    elif arg_type == 'quarter':
        arg_values = _parse_quarter_args(arg_range, start_date, end_date, reversed_par_seq)
    elif arg_type == 'month':
        arg_values = _parse_month_args(arg_range, start_date, end_date, reversed_par_seq)
    elif arg_type == 'table_index':
        arg_values = _parse_table_index_args(arg_range, symbols, allowed_code_suffix, reversed_par_seq)
    elif arg_type == 'none':
        arg_values = []
    else:
        raise ValueError('unexpected arg type:', arg_type)

    # build the args dict
    if (arg_name is None) and (additional_start_end.lower() != 'y'):
        kwargs = {}
    elif (arg_name is None) and (additional_start_end.lower() == 'y'):
        additional_args = _parse_additional_time_args(start_end_chunk_size, start_date, end_date)
        kwargs = ({**add_arg} for add_arg in additional_args)
    elif additional_start_end.lower() != 'y':
        # only standard args
        kwargs = ({arg_name: val} for val in arg_values)
    elif additional_start_end.lower() == 'y':
        # build additional start/end args
        additional_args = _parse_additional_time_args(start_end_chunk_size, start_date, end_date)
        import itertools
        kwargs = ({arg_name: val, **add_arg} for val, add_arg in
                  itertools.product(arg_values, additional_args))
    else:
        raise ValueError('unexpected additional_start_end:', additional_start_end)

    return kwargs


def fetch_batched_table_data(
        *,
        table: str,
        channel: str,
        arg_list: Any,
        parallel: bool = True,
        process_count: int = None,
        logger: logging.Logger = None,
        download_batch_size: int = 0,
        download_batch_interval: int = 0.,
) -> Generator[dict[str, Any], Any, None]:
    """ 一个Generator，顺序循环批量获取同一张数据表的数据，支持并行下载并逐个返回数据

    Parameters
    ----------
    table: str,
        数据表名，必须是database中定义的数据表
    channel: str,
        数据获取渠道，指定本地文件、金融数据API，或直接给出local_df，支持以下选项：
        - 'tushare'     : 从Tushare API获取金融数据，请自行申请相应权限和积分
        - 'akshare'     : 从AKshare API获取金融数据
        - 'eastmoney'   : 从东方财富网获取金融数据
    arg_list: iterable
        用于下载数据的函数参数
    parallel: bool, default True
        是否并行下载数据
    process_count: int, default None
        并行下载数据时，使用的进程数, 默认为None，表示使用cpu_count()个进程
    logger: logger
        用于记录下载数据的日志
    download_batch_size: int
        为降低网络请求的频率，在暂停之前连续网络请求的次数，单位为次，如果设置为0，则不暂停
    download_batch_interval: float
        为降低网络请求的频率，连续网络请求一定次数后，暂停的时间，单位为秒

    Yields
    -------
    dict: {'kwargs': kwargs, 'data': pd.DataFrame}
        下载的数据，包含参数和数据
    """

    fetch_table_data = _get_fetch_table_func(channel)

    # 如果当总下载量小于batch_size时，就不用暂停了(为了实现truncate，必须把arg_list转化为list)
    if not isinstance(arg_list, list):
        arg_list = list(arg_list)
    if len(arg_list) < download_batch_size:
        download_batch_interval = 0

    completed = 0
    if not parallel:
        for kwargs in arg_list:
            completed += 1
            df = fetch_table_data(table, **kwargs)
            if (download_batch_interval != 0) and (completed % download_batch_size == 0):
                time.sleep(download_batch_interval)
            if logger is not None:
                logger.info(f'[{table}:{kwargs}] {len(df)} rows downloaded')
            yield {'kwargs': kwargs, 'data': df}

    else:  # parallel
        # 使用ThreadPoolExecutor循环下载数据
        with ThreadPoolExecutor(max_workers=process_count) as worker:
            # 在parallel模式下，下载线程的提交和返回是分开进行的，为了实现分批下载，必须分批提交，提交一批
            # 数据后，等待结果返回，再提交下一批，因此，需要将arg_list分段，提交完一个batch之后，返回结果，
            # 再暂停，暂停后再继续提交
            submitted = 0
            # 将arg_list分段，每次下载batch_size个数据
            if download_batch_size == 0:
                arg_list_chunks = [arg_list]
            else:
                arg_list_chunks = list_truncate(arg_list, download_batch_size, as_list=False)

            for arg_sub_list in arg_list_chunks:
                futures = {}
                for kw in arg_sub_list:
                    futures.update({worker.submit(fetch_table_data, table, **kw): kw})
                    submitted += 1
                for f in as_completed(futures):
                    kwargs = futures[f]
                    completed += 1
                    if logger is not None:
                        logger.info(f'[{table}:{kwargs}] {len(f.result())} rows downloaded')
                    yield {'kwargs': kwargs, 'data': f.result()}

                if download_batch_interval != 0:
                    time.sleep(download_batch_interval)


def fetch_real_time_klines(
        *,
        channel: str,
        qt_codes: Union[str, list[str]],
        freq: str = 'd',
        parallel: bool = True,
        time_zone: str = 'local',
        verbose: bool = True,
        matured_kline_only: bool = False,
        logger: Any = None,
) -> pd.DataFrame:
    """ 从 channels 调用实时K线接口获取当天的最新实时K线数据，K线频率最低为‘d'，最高为'1min'
    获取的数据仅包括当天的数据，如果当天不是交易日，返回空数据框。

    Parameters
    ----------
    channel: str,
        数据获取渠道，指定本地文件、金融数据API，或直接给出local_df，支持以下选项：
        - 'tushare'     : 从Tushare API获取金融数据，请自行申请相应权限和积分
        - 'akshare'     : 从AKshare API获取金融数据
        - 'eastmoney'   : 从东方财富网获取金融数据
    qt_codes: str or [str],
        股票代码
    freq: str,
        数据频率，支持以下频率：
        - 'd': 如果
        - 'h':
        - '30min':
        - '15min':
        - '5min':
        - '1min':
    parallel: bool, optional, default True
        是否并行获取数据，默认True
    time_zone: str, optional
        时区，默认local即系统当地时区，可以指定特定的时区以获取不同时区市场的实时价格
    verbose: bool, optional, default False
        如果为True，给出更多的数据，否则返回简化的数据：
        - verbose = True: 'trade_time, symbol, name, pre_close, open, close, high, low, vol, amount'
        - verbose = False: 'trade_time, symbol, open, close, high, low, vol, amount'
    matured_kline_only: bool, default False
        是否只返回已经成熟的(也就是完整的上一个)K线数据，如果为True，则只返回已经成熟的K线数据，即已经完整的上一根K线数据
        如果为False，则返回最新的K线数据，哪怕这根K线数据并未完整
    logger: logger
        用于记录下载数据的日志

    Returns
    -------
    pd.DataFrame
    """

    fetch_realtime_kline = _get_realtime_kline_func(channel)

    data = []
    if isinstance(qt_codes, str):
        qt_codes = str_to_list(qt_codes)

    # 获取当前时间以及当天日期
    current_time = get_current_timezone_datetime(time_zone)
    today =current_time.strftime('%Y%m%d')

    # 使用ProcessPoolExecutor, as_completed加速数据获取，当parallel=False时，不使用多进程
    if parallel:
        with ThreadPoolExecutor(max_workers=10) as executor:
            futures = {
                executor.submit(fetch_realtime_kline, qt_code=symbol, date=today, freq=freq): symbol
                for symbol
                in qt_codes
            }
            for future in as_completed(futures):
                try:
                    df = future.result(timeout=2)
                    symbol = futures[future]
                except TimeoutError:
                    continue
                except Exception as exc:
                    print(f'Encountered an exception: {exc}')
                    if '权限' in str(exc):
                        continue
                    elif 'not_implemented' in str(exc):
                        continue
                    else:
                        raise exc
                else:
                    if df.empty:
                        continue
                    df['ts_code'] = symbol
                    # 根据当前时间确定哪个是matured kline，而不是直接取最后一个，因为最后一个可能是不完整的
                    if matured_kline_only:
                        last_matured_index = np.searchsorted(df.index, current_time, side='right') - 1
                        k_line = df.iloc[last_matured_index:last_matured_index + 1, :]
                    else:  # 否则就直接取最后一个，不管是否完整
                        k_line = df.iloc[-1:, :]
                    data.append(k_line)
    else:  # parallel == False, 不使用多进程
        for symbol in qt_codes:
            df = fetch_realtime_kline(qt_code=symbol, date=today, freq=freq)
            if df.empty:
                continue
            df['ts_code'] = symbol
            # 根据当前时间确定哪个是matured kline，而不是直接取最后一个，因为最后一个可能是不完整的
            if matured_kline_only:
                last_matured_index = np.searchsorted(df.index, current_time, side='right') - 1
                k_line = df.iloc[last_matured_index:last_matured_index + 1, :]
            else:  # 否则就直接取最后一个，不管是否完整
                k_line = df.iloc[-1:, :]
            data.append(k_line)
    try:
        data = pd.concat(data)
    except:
        return pd.DataFrame()  # 返回空DataFrame

    data = scrub_realtime_klines(data, verbose=verbose)

    return data


def fetch_real_time_quotes(
        *,
        channel: str,
        shares: str or [str],
        parallel: bool = True,
        logger: any = None,
) -> pd.DataFrame:
    """ 从 channels 调用实时盘口API获取实时盘口数据，包括实时价格、成交数据、五档委买委卖数据等。

    Parameters
    ----------
    channel: str,
        数据获取渠道，指定本地文件、金融数据API，或直接给出local_df，支持以下选项：
        - 'tushare'     : 从Tushare API获取金融数据，请自行申请相应权限和积分
        - 'akshare'     : 从AKshare API获取金融数据
        - 'eastmoney'   : 从东方财富网获取金融数据
    shares: str or [str],
        股票代码
    parallel: bool, optional, default True
        是否并行获取数据，默认True
    logger: logger
        用于记录下载数据的日志

    Returns
    -------
    pd.DataFrame
    """
    raise NotImplementedError


def scrub_table_data(data: pd.DataFrame, table: str) -> pd.DataFrame:
    """ 清洗数据，确保数据data的字段和数据类型符合数据表table的定义

    Parameters
    ----------
    data: pd.DataFrame,
        下载的数据
    table: str,
        数据表名，必须是API_MAP中定义的数据表

    Returns
    -------
    pd.DataFrame:
        清洗后的数据
    """
    # TODO: this function is now not utilized; data scrubbing is done in datasource
    #  but later data scrubbing shoule be done here, and datasource only does basic checks
    #  following works should be done:
    #  1, make sure data column numbers match table schema
    #  2, remove all NaN data rows
    #  3, reindex data column names
    #  4, make DataFrame index according to table prime keys and remove duplicate indexes
    #  5, check data types, make sure all data types fit table schema
    #  6, cut off strings that exceeds varchar limit
    raise NotImplementedError


def scrub_realtime_quote_data(raw_data, verbose) -> pd.DataFrame:
    """ 清洗数据，去除不一致及错误数据，并使数据符合realtime_quote的数据格式"""
    raise NotImplementedError


def scrub_realtime_klines(raw_data, verbose) -> pd.DataFrame:
    """ 清洗数据，去除不一致及错误的数据，并使数据符合实时K线图的数据格式"""

    if "trade_time" not in raw_data.columns:
        # trade_time 在 index 中，将它移到frame中
        assert raw_data.index.name in ['trade_time', 'trade_date'], AssertionError('trade_time should be in index')
        raw_data['trade_time'] = raw_data.index
    # 重新整理数据列名称
    if verbose:
        target_columns = ['trade_time', 'ts_code', 'name', 'pre_close', 'open', 'close',
                          'high', 'low', 'vol', 'amount']

    else:

        target_columns = ['trade_time', 'ts_code', 'open', 'close', 'high', 'low', 'vol', 'amount']

    data = raw_data.reindex(
            columns=target_columns
    )
    # set index
    data.set_index('trade_time', inplace=True)

    return data


# ======================================
# 一系列函数数据获取原子函数，针对不同的API主参数类型，生成下载数据的参数序列
# ======================================
def _parse_list_args(arg_range: str or [str], list_arg_filter: str or [str] = None, reversed_par_seq: bool = False):
    """ 解析list类型的参数，生成下载数据的参数序列

    Parameters
    ----------
    arg_range: str or list of str,
        用于下载数据时的筛选参数，某些数据表以列表的形式给出可筛选参数，如stock_basic表，它有一个可筛选
        参数"exchange"，选项包含 'SSE', 'SZSE', 'BSE'，可以通过此参数限定下载数据的范围。
    list_arg_filter: str or list of str, optional
        用于下载数据时的筛选参数，某些数据表以列表的形式给出可筛选参数，如stock_basic表，它有一个可筛选
        参数"exchange"，选项包含 'SSE', 'SZSE', 'BSE'，可以通过此参数限定下载数据的范围。
        如果filter_arg为None，则下载所有数据。
        例如，下载stock_basic表数据时，下载以下输入均为合法输入：
        - 'SZSE'
            仅下载深圳交易所的股票数据
        - ['SSE', 'SZSE']
        - 'SSE, SZSE'
            上面两种写法等效，下载上海和深圳交易所的股票数据
    reversed_par_seq: bool, default False
        是否将参数序列反转，如果为True，则会将参数序列反转，用于下载数据时的优化

    Returns
    -------
    list generator:
        用于下载数据的参数序列
    """
    if isinstance(arg_range, str):
        arg_range = str_to_list(arg_range)

    if reversed_par_seq:
        arg_range = arg_range[::-1]

    if list_arg_filter is None:
        return (item for item in arg_range)

    if (isinstance(list_arg_filter, str)) and ":" not in list_arg_filter:
        # 如果list_arg_filter是字符串，且不包含":"，则将其转换为列表
        list_arg_filter = str_to_list(list_arg_filter)

    if isinstance(list_arg_filter, str):
        # 如果list_arg_filter是字符串，且包含":"，则将其转换为上下限
        list_arg_bounds = list_arg_filter.split(':')
        list_idx_lower = arg_range.index(list_arg_bounds[0])
        list_idx_upper = arg_range.index(list_arg_bounds[1])
        if list_idx_lower > list_idx_upper:
            # reverse the lower and upper index if lower is larger than upper
            list_idx_lower, list_idx_upper = list_idx_upper, list_idx_lower
        # 确认上下限的位置
        return (item for item in arg_range[list_idx_lower:list_idx_upper + 1])

    return (item for item in arg_range if item in list_arg_filter)


def _parse_datetime_args(arg_range: str, start_date: str, end_date: str,
                         freq: str = 'd', reversed_par_seq: bool = False):
    """ 根据开始和结束日期，生成数据获取的参数序列

    Parameters
    ----------
    arg_range: str,
        表示数据的时间范围的起始日期，如'20210101'表示从2021年1月1日开始
    start_date: str, YYYYMMDD
        数据下载的开始日期
    end_date: str, YYYYMMDD
        数据下载的结束日期
    freq: str, optional
        数据下载的频率，支持以下选项：
        - 'd':  日频率
        - 'w':  周频率
        - 'm':  月频率
    reversed_par_seq: bool, default False
        是否将参数序列反转，如果为True，则会将参数序列反转，用于下载数据时的优化

    Returns
    -------
    list:
        用于下载数据的参数序列
    """

    start_date, end_date = _ensure_date_sequence(arg_range, start_date, end_date)

    if freq is None:
        freq = 'd'
    if freq.lower() == 'w':
        freq = 'w-Fri'
    if freq.lower() == 'm':
        freq = 'ME'

    res = pd.date_range(start_date, end_date, freq=freq).strftime('%Y%m%d').to_list()

    if reversed_par_seq:
        return res[::-1]
    else:
        return res


def _parse_trade_date_args(arg_range: str, start_date: str, end_date: str,
                           freq: str = 'd', market: str = 'SSE', reversed_par_seq: bool = False):
    """ 根据开始和结束日期，生成数据获取的参数序列

    Parameters
    ----------
    start_date: str,
        数据下载的开始日期
    end_date: str,
        数据下载的结束日期
    freq: str, optional
        数据下载的频率，支持以下选项：
        - 'd':  日频率
        - 'w':  周频率
        - 'm':  月频率
    market: str, optional
        交易市场，支持数据表Trade_calendar中定义的交易市场
    reversed_par_seq: bool, default False
        是否将参数序列反转，如果为True，则会将参数序列反转，用于下载数据时的优化

    Returns
    -------
    list:
        用于下载数据的参数序列
    """

    from .utilfuncs import is_market_trade_day, nearest_market_trade_day
    full_dates = _parse_datetime_args(arg_range, start_date, end_date, freq, reversed_par_seq)

    if freq is None:
        freq = 'd'
    # 如果频率是月或周，分别查找所有日期的nearest_trade_day并删除重复日期
    if freq.lower() in ['w', 'm']:
        trade_dates = list(map(nearest_market_trade_day, full_dates))
        trade_dates = list(set(trade_dates))
        trade_dates.sort(reverse=reversed_par_seq)
        trade_dates = [date.strftime('%Y%m%d') for date in trade_dates]
    else:
        # 如果频率是日，删除所有非交易日
        trade_dates = [date for date in full_dates if is_market_trade_day(date, market)]

    return trade_dates


def _parse_table_index_args(arg_range: str, symbols: str, allowed_code_suffix: str = None,
                            reversed_par_seq: bool = False):
    """ 根据开始和结束日期，生成数据获取的参数序列

    Parameters
    ----------
    arg_range: str,
        数据表名，定义为basics的数据表，包含股票代码等基本信息
    symbols: str or list of str,
        用于下载数据的股票代码，如果给出了symbols，只有这些股票代码的数据会被下载
    allowed_code_suffix: str, optional
        用于下载数据的股票代码后缀，如果给出了allowed_code_suffix，只有这些股票代码的数据会被下载
    reversed_par_seq: bool, default False
        是否将参数序列反转，如果为True，则会将参数序列反转，用于下载数据时的优化

    Returns
    -------
    list:
        用于下载数据的参数序列
    """

    from qteasy import QT_DATA_SOURCE

    df_s, df_i, df_f, df_ft, df_o, df_ths = QT_DATA_SOURCE.get_all_basic_table_data()

    table_name = arg_range

    if table_name == 'stock_basic':
        all_args = df_s.index.to_list()
    elif table_name == 'index_basic':
        all_args = df_i.index.to_list()
    elif table_name == 'fund_basic':
        all_args = df_f.index.to_list()
    elif table_name == 'future_basic':
        all_args = df_ft.index.to_list()
    elif table_name == 'opt_basic':
        all_args = df_o.index.to_list()
    elif table_name == 'ths_index_basic':
        all_args = df_ths.index.to_list()
    else:
        raise ValueError(f'unknown table name {table_name}')

    if symbols is not None:  # assert symbols is a str, 进行第一次筛选
        # 冒号分隔的字符串，表示股票代码的上下限
        if ':' in symbols:
            symbols = symbols.split(':')
            lower = symbols[0].split('.')[0]
            upper = symbols[1].split('.')[0]
            if lower > upper:
                lower, upper = upper, lower
            all_args = [arg for arg in all_args if (lower <= arg.split('.')[0] <= upper)]

        else:
            symbols = str_to_list(symbols)
            all_args = [arg for arg in all_args if arg in symbols]

    if allowed_code_suffix:  # assert allowed_code_suffix is a str, 进行第二次筛选
        suffix = str_to_list(allowed_code_suffix)
        all_args = [arg for arg in all_args if arg[-2:] in suffix]

    if reversed_par_seq:
        all_args = all_args[::-1]

    return all_args


def _parse_quarter_args(arg_range: str, start_date: str, end_date: str, reversed_par_seq: bool = False) -> list:
    """ 根据开始和结束日期，生成数据获取的参数序列，类似2022Q1等

    Parameters
    ----------
    arg_range: str,
        可以下载的最早数据的季度标识，如1976Q1标识1976年第一季度
    start_date: str,
        数据下载的开始日期
    end_date: str,
        数据下载的结束日期

    Returns
    -------
    list:
        用于下载数据的参数序列
    """
    start_date, end_date = _ensure_date_sequence(arg_range, start_date, end_date)

    # calculate absolute quarter number
    start_quarter = _convert_date_to_absolute_quarter(start_date)
    end_quarter = _convert_date_to_absolute_quarter(end_date)

    # convert year and quarter to string
    quarter_list = list(range(start_quarter, end_quarter + 1))
    quarter_list = [_convert_absolute_quarter_to_quarter_str(q) for q in quarter_list]

    if reversed_par_seq:
        return quarter_list[::-1]

    return quarter_list


def _get_year_and_quarter(date) -> tuple:
    """ convert date like 2020-01-01 to year and quarter like 2020, 1"""
    return date.year, (date.month - 1) // 3 + 1


def _get_year_and_month(date) -> tuple:
    """ convert date like 2020-03-01 to year and month like 2020, 3"""
    return date.year, date.month


def _convert_date_to_absolute_quarter(date) -> int:
    """ convert date like 2020-01-01 to absolute quarter number like 2020*4+1"""
    year, quarter = _get_year_and_quarter(date)
    return year * 4 + quarter


def _convert_quarter_str_to_absolute_quarter(quarter: str) -> int:
    """ converte quarter str like 2020Q1 to absolute quarter number like 2020*4+1"""
    year, quarter = quarter.split('Q')
    return int(year) * 4 + int(quarter)


def _convert_date_to_absolute_month(date) -> int:
    """ convert date like 2020-03-01 to absolute month number like 2020*12+3"""
    year, month = _get_year_and_month(date)
    return year * 12 + month


def _convert_month_str_to_absolute_month(date: str) -> int:
    """ convert month str like 202003 to absolute month like 2020*12+3"""
    year, month = date[:4], date[4:]
    return int(year) * 12 + int(month)


def _convert_absolute_quarter_to_quarter_str(quarter_num) -> str:
    """ convert absolute quarter number like 2020*4+1 to quarter like 2020Q1"""
    year, quarter = quarter_num // 4, quarter_num % 4
    if quarter == 0:
        year = year - 1
        quarter = 4
    return f'{year}Q{quarter}'


def _convert_absolute_month_to_month_str(month_num) -> str:
    """ convert absolute month number like 2020*12+3 to month like 202003"""
    year, month = month_num // 12, month_num % 12
    if month == 0:
        year = year - 1
        month = 12
    return f'{year}{month:02d}'


def _parse_month_args(arg_range: str, start_date: str, end_date: str, reversed_par_seq: bool = False) -> list:
    """ 根据开始和结束日期，生成数据获取的参数序列

    Parameters
    ----------
    arg_range: str,
        数据下载的开始日期
    start_date: str,
        数据下载的开始日期
    end_date: str,
        数据下载的结束日期
    reversed_par_seq: bool, default False
        是否将参数序列反转，如果为True，则会将参数序列反转，用于下载数据时的优化

    Returns
    -------
    list:
        用于下载数据的参数序列
    """
    start_date, end_date = _ensure_date_sequence(arg_range, start_date, end_date)

    # calculate absolute month number
    start_month = _convert_date_to_absolute_month(start_date)
    end_month = _convert_date_to_absolute_month(end_date)

    # complete list with all months
    month_list = list(range(start_month, end_month + 1))

    # convert year and month to string
    month_list = [_convert_absolute_month_to_month_str(month) for month in month_list]

    if reversed_par_seq:
        return month_list[::-1]

    return month_list


def _parse_additional_time_args(chunk_size, start_date, end_date) -> list:
    """ 生成附加的开始/结束日期参数。对于分钟类型的数据表，按照qt_code生成参数序列
    还有可能无法完整下载所有数据，因此需要增加start/end限制参数，限制每次下载的数据量
    确保数据下载的完整性。

    Parameters
    ----------
    chunk_size: str,
        分批下载数据时，每一批数据包含的数据量，单位为天，表示每次下载多少天的数据
    start_date: str,
        数据下载的开始日期
    end_date: str,
        数据下载的结束日期

    Returns
    -------
    list: [{'start': date, 'end': date}, ...]
        用于下载数据的额外参数序列，在主参数序列的基础上，增加了start和end参数
    """

    start_date, end_date = _ensure_date_sequence('19700101', start_date, end_date)

    if chunk_size is None:
        return [{'start': start_date.strftime('%Y%m%d'), 'end': end_date.strftime('%Y%m%d')}]
    try:
        chunk_size = int(chunk_size)
    except ValueError:
        chunk_size = 0

    if chunk_size <= 0:
        return [{'start': start_date.strftime('%Y%m%d'), 'end': end_date.strftime('%Y%m%d')}]  # return the whole period

    start_end_chunk_lbounds = list(pd.date_range(start=start_date,
                                                 end=end_date,
                                                 freq=f'{chunk_size}d'
                                                 ).strftime('%Y%m%d'))
    start_end_chunk_rbounds = start_end_chunk_lbounds[1:]

    start_end_chunk_rbounds.append(end_date.strftime('%Y%m%d'))
    chunked_additional_args = [{'start': s, 'end': e} for s, e in
                               zip(start_end_chunk_lbounds, start_end_chunk_rbounds)]

    return chunked_additional_args


def _ensure_date_sequence(first_date, start_date, end_date) -> tuple:
    """ 确保开始和结束日期在first_date之后，如果不是，则交换开始和结束日期

    Parameters
    ----------
    first_date: str,
        数据表中的最早日期
    start_date: str,
        数据下载的开始日期
    end_date: str,
        数据下载的结束日期

    Returns
    -------
    tuple: start_date, end_date
        确保开始和结束日期在first_date之后的开始和结束日期
    """

    first_date = pd.to_datetime(first_date)
    start_date = pd.to_datetime(start_date)
    end_date = pd.to_datetime(end_date)

    if start_date < first_date:
        start_date = first_date
    if end_date < first_date:
        end_date = first_date
    if end_date < start_date:
        start_date, end_date = end_date, start_date

    return start_date, end_date


def get_dependent_table(table: str, channel: str) -> str or None:
    """ 获取数据表的依赖表. 依赖表是指在获取某个数据表之前，需要先获取的数据表

    依赖表主要包括在api_map中定义的fill_arg_type为'table_index'的数据表，如stock_basic表

    Parameters
    ----------
    table: str,
        数据表名，必须是database中定义的数据表
    channel: str,
        数据获取渠道，指定本地文件、金融数据API，或直接给出local_df，支持以下选项：
        - 'tushare'     : 从Tushare API获取金融数据，请自行申请相应权限和积分
        - 'akshare'     : 从AKshare API获取金融数据
        - 'eastmoney'   : 从东方财富网获取金融数据

    Returns
    -------
    str:
        数据表的依赖表名称

    Notes
    -----
    如果数据表不存在或者无法找到依赖表，则返回None
    """

    api_map = get_api_map(channel=channel)

    if table not in api_map.index:
        return None

    cur_table = api_map.loc[table]
    fill_type = cur_table.fill_arg_type
    if fill_type == 'trade_date':
        return 'trade_calendar'
    elif fill_type == 'table_index':
        return cur_table.arg_rng


@lru_cache(maxsize=4)
def get_api_map(channel: str) -> pd.DataFrame:
    """ 获取指定金融数据API的MAP表

    Parameters
    ----------
    channel: str,
        数据获取渠道，金融数据API，支持以下选项:
        - 'tushare'     : 从Tushare API获取金融数据，请自行申请相应权限和积分
        - 'akshare'     : 从AKshare API获取金融数据
        - 'eastmoney'   : 从东方财富网获取金融数据

    Returns
    -------
    pd.DataFrame
        根据channel返回对应的API MAP表，DataFrame格式
    """

    if channel == 'tushare':
        API_MAP = TUSHARE_API_MAP
        MAP_COLUMNS = API_MAP_COLUMNS
    elif channel == 'akshare':
        API_MAP = AKSHARE_API_MAP
        MAP_COLUMNS = API_MAP_COLUMNS
    elif channel == 'eastmoney':
        API_MAP = EASTMONEY_API_MAP
        MAP_COLUMNS = API_MAP_COLUMNS
    else:
        raise NotImplementedError(f'channel {channel} is not supported')

    api_map = pd.DataFrame(API_MAP).T
    api_map.columns = MAP_COLUMNS

    return api_map


"""
TUSHARE API MAP表column的定义：
1, api:                     对应的tushare API函数名, 参见tsfuncs.py中的定义

2, fill_arg_name:           从tushare下载数据时的关键参数，使用该关键参数来分多次下载完整的数据
                            例如，所有股票所有日期的日K线数据一共有超过1500万行，这些数据无法一次性下载
                            必须分步多次下载。要么每次下载一只股票所有日期的数据，要么下载每天所有股票的数据。
                            -   如果每次下载一只股票的数据，股票的代码就是关键参数，将所有股票的代码依次作为关键参数
                                输入tushare API，就可以下载所有数据
                            -   如果每次下载一天的数据，交易日期就是关键参数，将所有的交易日作为关键参数，也可以下载
                                所有的数据
                            对很多API来说，上述两种方式都是可行的，但是以日期为关键参数的方式更好，因为两个原因：
                            1,  作为关键参数，交易日的数量更少，每年200个交易日，20年的数据也只需要下载4000次，如果
                                使用股票代码，很多证券类型的代码远远超过4000个
                            2,  补充下载数据时更方便，如果过去20年的数据都已经下载好了，仅需要补充下载最新一天的数据
                                如果使用日期作为关键参数，仅下载一次就可以了，如果使用证券代码作为关键参数，需要下载
                                数千次
                            因此，应该尽可能使用日期作为关键参数
                            可惜的是，并不是所有API都支持使用日期作为关键参数，所以，只要能用日期的数据表，都用日期
                            为关键参数，否则采用其他类型的关键参数。

3, fill_arg_type:           关键参数的数据类型，可以为list、table_index、datetime、trade_date，含义分别为：
                            -   list: 
                                列表类型，这个关键参数的取值只能是列表中的某一个元素
                            -   table_index: 
                                表索引，这个关键参数的取值只能是另一张数据表的索引值，这通常表示tushare不支持使用日期
                                作为关键参数，因此必须使用一系列股票或证券代码作为关键参数的值，这些值保存在另一张数据
                                表中，且是这个表的索引值。例如，所有股票的代码就保存在stock_basic表的索引中。
                            -   datetime:
                                日期类型，表示使用日期作为下载数据的关键参数。此时的日期不带时间类型，包含交易日与非
                                交易日
                            -   trade_date:
                                交易日，与日期类型功能相同，但作为关键参数的日期必须为交易日

4, arg_rng:                 关键参数的取值范围，
                            -   如果数据类型为datetime或trade_date，是一个起始日期，表示取值范围从该日期到今日之间
                            -   如果数据类型为table_index，取值范围是一张表，如stock_basic，表示数据从stock_basic的
                                索引中取值
                            -   如果数据类型为list，则直接给出取值范围，如"SSE,SZSE"表示可以取值SSE以及SZSE。

5, arg_allowed_code_suffix:
                            table_index类型取值范围的限制值，限制只有特定后缀的证券代码才会被用作参数下载数据。
                            例如，取值"SH,SZ"表示只有以SH、SZ结尾的证券代码才会被用作参数从tushare下载数据。

6, arg_allow_start_end:     使用table_index类型参数时，是否同时允许传入开始结束日期作为参数。如果设置为"Y"，则会在使用
                            table_index中的代码作为参数下载数据时，同时传入开始和结束日期作为附加参数，否则仅传入代码

7, start_end_chunk_size:    传入开始结束日期作为附加参数时，是否分块下载。可以设置一个正整数或空字符串如"300"。如果设置了
                            一个正整数字符串，表示一个天数，并将开始结束日期之间的数据分块下载，每个块中数据的时间跨度不超
                            过这个天数。
                            例如，设置该参数为100，则每个分块内的时间跨度不超过100天
"""

API_MAP_COLUMNS = [
    'api',  # 1, 从channel获取数据时使用的api名
    'fill_arg_name',  # 2, 获取数据时使用的api参数名
    'fill_arg_type',  # 3, 从获取数据时使用的api参数类型
    'arg_rng',  # 4, 获取数据时使用的api参数取值范围
    'allowed_code_suffix',  # 5, 获取数据时使用的api参数允许的股票代码后缀
    'allow_start_end',  # 6, 获取数据时使用的api参数需要的额外参数类型
    'start_end_chunk_size',  # 7, 从tushare获取数据时使用的api参数start_date和end_date时的分段大小
    # TODO: 修改column6/column7 for minor upgrades of version 1.4：
    #  将6改为"table_row_limiter”,含义是限制同时下载的数据行方法，取值包括：
    #  -   '':   没有额外的参数用于限制下载数据的行数
    #  -   'Y':  使用开始结束日期作为参数，限制下载数据的行数
    #  -   'C':  使用limit/offset作为参数，直接限制下载数据的行数
    #  -   'other':  其他方法，如使用日期作为参数，但不限制下载数据的行数
    #  将7改为“table_row_limiter_args”，含义是限制下载数据行数的参数，取值包括：
    #  -   当table_row_limiter为''时，table_row_limiter_args为''
    #  -   当table_row_limiter为'Y'时，table_row_limiter_args为起止日期之间的天数，如200代表最多下载200天的数据
    #  -   当table_row_limiter为'C'时，table_row_limiter_args为limit/offset的值，如100代表每次下载100行数据
]

TUSHARE_API_MAP = {
    'trade_calendar':
        ['trade_cal', 'exchange', 'list', 'SSE, SZSE, CFFEX, SHFE, CZCE, DCE, INE', '', '', ''],

    # 'hk_trade_calendar':  # tsfuncs
    #     ['hk_tradecal', 'none', 'none', 'none', '', 'C', '2000'],

    # 'us_trade_calendar':  # tsfuncs
    #     ['us_tradecal', 'none', 'none', 'none', '', 'C', '6000'],

    'stock_basic':
        ['stock_basic', 'exchange', 'list', 'SSE,SZSE,BSE', '', '', '', ],

    # 'hk_stock_basic':  # tsfuncs
    #     ['hk_stock_basic', 'none', 'none', 'none', '', '', ''],
    #
    # 'us_stock_basic':  # tsfuncs
    #     ['us_stock_basic', 'none', 'none', 'none', '', 'C', '6000'],

    'stock_names':
        ['namechange', 'ts_code', 'table_index', 'stock_basic', '', 'Y', '300'],

    'stock_company':
        ['stock_company', 'exchange', 'list', 'SSE, SZSE, BSE', '', '', ''],

    'stk_managers':
        ['stk_managers', 'ann_date', 'datetime', '19901211', '', '', ''],

    'new_share':
        ['new_share', 'none', 'none', 'none', '', 'Y', '200'],

    'money_flow':
        ['moneyflow', 'trade_date', 'trade_date', '20000101', '', '', ''],

    'stock_limit':  # start date verified
        ['stk_limit', 'trade_date', 'trade_date', '20070101', '', '', ''],

    'stock_suspend':
        ['suspend_d', 'trade_date', 'trade_date', '20000101', '', '', ''],

    'hs_money_flow':
        ['moneyflow_hsgt', 'trade_date', 'trade_date', '20100101', '', '', ''],

    'hs_top10_stock':
        ['hsgt_top10', 'trade_date', 'trade_date', '20100101', '', '', ''],

    'hk_top10_stock':
        ['ggt_top10', 'trade_date', 'trade_date', '20100101', '', '', ''],

    'index_basic':
        ['index_basic', 'market', 'list', 'SSE,MSCI,CSI,SZSE,CICC,SW,OTH', '', '', ''],

    'fund_basic':
        ['fund_basic', 'market', 'list', 'E,O', '', '', ''],

    'future_basic':
        ['future_basic', 'exchange', 'list', 'CFFEX,DCE,CZCE,SHFE,INE', '', '', ''],

    'opt_basic':
        ['options_basic', 'exchange', 'list', 'SSE,SZSE,CFFEX,DCE,CZCE,SHFE', '', '', ''],

    'ths_index_basic':
        ['ths_index', 'exchange', 'list', 'A, HK, US', '', '', ''],

    'sw_industry_basic':
        ['index_classify', 'src', 'list', 'sw2014, sw2021', '', '', ''],

    'stock_1min':
        ['mins1', 'ts_code', 'table_index', 'stock_basic', '', 'y', '30'],

    'stock_5min':
        ['mins5', 'ts_code', 'table_index', 'stock_basic', '', 'y', '90'],

    'stock_15min':
        ['mins15', 'ts_code', 'table_index', 'stock_basic', '', 'y', '180'],

    'stock_30min':
        ['mins30', 'ts_code', 'table_index', 'stock_basic', '', 'y', '360'],

    'stock_hourly':
        ['mins60', 'ts_code', 'table_index', 'stock_basic', '', 'y', '360'],

    'stock_daily':
        ['daily', 'trade_date', 'trade_date', '19901211', '', '', ''],

    'stock_weekly':
        ['weekly', 'trade_date', 'trade_date', '19901221', '', '', ''],

    'stock_monthly':
        ['monthly', 'trade_date', 'trade_date', '19901211', '', '', ''],

    'hk_stock_daily':
        ['hk_daily', 'trade_date', 'hk_trade_date', '19901211', '', '', ''],

    'us_stock_daily':
        ['us_daily', 'trade_date', 'us_trade_date', '19601211', '', '', ''],

    'index_1min':
        ['mins1', 'ts_code', 'table_index', 'index_basic', 'SH,SZ', 'y', '30'],

    'index_5min':
        ['mins5', 'ts_code', 'table_index', 'index_basic', 'SH,SZ', 'y', '90'],

    'index_15min':
        ['mins15', 'ts_code', 'table_index', 'index_basic', 'SH,SZ', 'y', '180'],

    'index_30min':
        ['mins30', 'ts_code', 'table_index', 'index_basic', 'SH,SZ', 'y', '360'],

    'index_hourly':
        ['mins60', 'ts_code', 'table_index', 'index_basic', 'SH,SZ', 'y', '360'],

    'index_daily':
        ['index_daily', 'ts_code', 'table_index', 'index_basic', 'SH,CSI,SZ', 'y', ''],

    'index_weekly':
        ['index_weekly', 'trade_date', 'trade_date', '19910705', '', '', ''],

    'index_monthly':
        ['index_monthly', 'trade_date', 'trade_date', '19910731', '', '', ''],

    'ths_index_daily':
        ['ths_daily', 'trade_date', 'trade_date', '20070801', '', '', ''],

    'ths_index_weight':
        ['ths_member', 'ts_code', 'table_index', 'ths_index_basic', '', '', ''],

    'ci_index_daily':
        ['ci_daily', 'trade_date', 'trade_date', '20100101', '', '', ''],

    'sw_index_daily':
        ['sw_daily', 'trade_date', 'trade_date', '20000101', '', '', ''],

    'global_index_daily':
        ['index_global', 'trade_date', 'trade_date', '19901211', '', '', ''],

    'fund_1min':
        ['mins1', 'ts_code', 'table_index', 'fund_basic', 'SH,SZ', 'y', '30'],

    'fund_5min':
        ['mins5', 'ts_code', 'table_index', 'fund_basic', 'SH,SZ', 'y', '90'],

    'fund_15min':
        ['mins15', 'ts_code', 'table_index', 'fund_basic', 'SH,SZ', 'y', '180'],

    'fund_30min':
        ['mins30', 'ts_code', 'table_index', 'fund_basic', 'SH,SZ', 'y', '360'],

    'fund_hourly':
        ['mins60', 'ts_code', 'table_index', 'fund_basic', 'SH,SZ', 'y', '360'],

    'fund_daily':
        ['fund_daily', 'trade_date', 'trade_date', '19980417', '', '', ''],

    'fund_nav':
        ['fund_net_value', 'nav_date', 'datetime', '20000107', '', '', ''],

    'fund_share':
        ['fund_share', 'ts_code', 'table_index', 'fund_basic', '', '', ''],

    'fund_manager':
        ['fund_manager', 'ts_code', 'table_index', 'fund_basic', 'OF, SZ, SH', '', ''],

    'future_mapping':
        ['fut_mapping', 'trade_date', 'trade_date', '19901219', '', '', ''],

    'future_1min':  # future_xmin 表应该通过trade_time(start/end)来切表索引，而不是通过table_index
        ['ft_mins1', 'ts_code', 'table_index', 'future_basic', '', 'y', '30'],

    'future_5min':
        ['ft_mins5', 'ts_code', 'table_index', 'future_basic', '', 'y', '90'],

    'future_15min':
        ['ft_mins15', 'ts_code', 'table_index', 'future_basic', '', 'y', '180'],

    'future_30min':
        ['ft_mins30', 'ts_code', 'table_index', 'future_basic', '', 'y', '360'],

    'future_hourly':
        ['ft_mins60', 'ts_code', 'table_index', 'future_basic', '', 'y', '360'],

    'future_daily':
        ['future_daily', 'trade_date', 'trade_date', '19950417', '', '', ''],

    'future_weekly':
        ['fut_weekly', 'trade_date', 'trade_date', '19950417', '', '', ''],

    'future_monthly':
        ['fut_monthly', 'trade_date', 'trade_date', '19950417', '', '', ''],

    'options_1min':  # options_xmin 表应该通过trade_time(start/end)来切表索引，而不是通过table_index
        ['mins1', 'ts_code', 'table_index', 'opt_basic', '', 'y', '30'],

    'options_5min':
        ['mins5', 'ts_code', 'table_index', 'opt_basic', '', 'y', '90'],

    'options_15min':
        ['mins15', 'ts_code', 'table_index', 'opt_basic', '', 'y', '180'],

    'options_30min':
        ['mins30', 'ts_code', 'table_index', 'opt_basic', '', 'y', '360'],

    'options_hourly':
        ['mins60', 'ts_code', 'table_index', 'opt_basic', '', 'y', '360'],

    'options_daily':
        ['options_daily', 'trade_date', 'trade_date', '20150209', '', '', ''],

    'stock_adj_factor':
        ['adj_factors', 'trade_date', 'trade_date', '19901219', '', '', ''],

    'fund_adj_factor':
        ['fund_adj', 'trade_date', 'trade_date', '19980407', '', '', ''],

    'stock_indicator':
        ['daily_basic', 'trade_date', 'trade_date', '19990101', '', '', ''],

    'stock_indicator2':
        ['bak_daily', 'trade_date', 'trade_date', '19990101', '', '', ''],

    'hk_stock_indicator':  # tsfuncs
        ['hk_indicators', 'trade_date', 'hk_trade_date', '19900101', '', '', ''],

    'us_stock_indicator':  # tsfuncs
        ['us_indicators', 'trade_date', 'us_trade_date', '19600101', '', '', ''],

    'index_indicator':
        ['index_dailybasic', 'trade_date', 'trade_date', '20040102', '', '', ''],

    'index_weight':
        ['composite', 'trade_date', 'datetime', '20050408', '', '', ''],

    'income':
        ['income', 'ts_code', 'table_index', 'stock_basic', '', 'Y', ''],

    'balance':
        ['balance', 'ts_code', 'table_index', 'stock_basic', '', 'Y', ''],

    'cashflow':
        ['cashflow', 'ts_code', 'table_index', 'stock_basic', '', 'Y', ''],

    'financial':
        ['indicators', 'ts_code', 'table_index', 'stock_basic', '', 'Y', ''],

    'forecast':
        ['forecast', 'ts_code', 'table_index', 'stock_basic', '', 'Y', ''],

    'express':
        ['express', 'ts_code', 'table_index', 'stock_basic', '', 'Y', ''],

    'dividend':
        ['dividend', 'ts_code', 'table_index', 'stock_basic', '', '', ''],

    'top_list':
        ['top_list', 'trade_date', 'trade_date', '20050101', '', '', ''],

    'top_inst':
        ['top_inst', 'trade_date', 'trade_date', '19901211', '', '', ''],

    'sw_industry_detail':
        ['index_member_all', 'l1_code', 'list',
         '801010.SI, 801020.SI, 801030.SI, 801040.SI, 801050.SI, 801080.SI, '
         '801110.SI, 801120.SI, 801130.SI, 801140.SI, 801150.SI, 801160.SI, '
         '801170.SI, 801180.SI, 801200.SI, 801210.SI, 801230.SI, 801710.SI, '
         '801720.SI, 801730.SI, 801740.SI, 801750.SI, 801760.SI, 801770.SI, '
         '801780.SI, 801790.SI, 801880.SI, 801890.SI, 801950.SI, 801960.SI, '
         '801970.SI, 801980.SI',
         '', '', ''],

    'block_trade':
        ['block_trade', 'trade_date', 'trade_date', '20100101', '', '', ''],

    'stock_holder_trade':
        ['stk_holdertrade', 'ann_date', 'datetime', '20100101', '', '', ''],

    'margin':
        ['margin', 'exchange_id', 'list', 'SSE,SZSE,BSE', '', '', ''],

    'margin_detail':
        ['margin_detail', 'trade_date', 'trade_date', '20190910', '', '', ''],

    'shibor':
        ['shibor', 'none', 'none', '', '', 'Y', '365'],

    'libor':
        ['libor', 'none', 'none', '', '', 'Y', '365'],

    'hibor':
        ['hibor', 'none', 'none', '', '', 'Y', '365'],

    'wz_index':
        ['wz_index', 'none', 'none', '', '', 'Y', '365'],

    'gz_index':
        ['gz_index', 'none', 'none', '', '', 'Y', '365'],

    'cn_gdp':
        ['cn_gdp', 'start', 'quarter', '1976Q4', '', '', ''],

    'cn_cpi':
        ['cn_cpi', 'start', 'month', '19760101', '', '', ''],

    'cn_ppi':
        ['cn_ppi', 'start', 'month', '19760101', '', '', ''],

    'cn_money':
        ['cn_money', 'start', 'month', '19760101', '', '', ''],

    'cn_sf':
        ['cn_sf', 'start', 'month', '19760101', '', '', ''],

    'cn_pmi':
        ['cn_pmi', 'start', 'month', '19760101', '', '', ''],
}

TUSHARE_REALTIME_API_MAP = {
    'realtime_bars':  # 实时行情数据
        ['realtime_min', 'ts_code', 'list', 'none', '', 'N', '', ''],
    'realtime_quotes':
        ['realtime_quote', 'ts_code', 'list', 'none', '', 'N', '', '']
}

AKSHARE_API_MAP = {

}

AKSHARE_REALTIME_API_MAP = {
    'realtime_bars':  # 实时行情数据
        ['not_implemented', 'symbols', 'list', 'none', '', 'N', '', ''],

    'realtime_quotes':
        ['not_implemented', 'symbols', 'list', 'none', '', 'N', '', '']
}

EASTMONEY_API_MAP = {  # 从EastMoney的数据API不区分asset_type，只要给出qt_code即可，因此index和stock共用API
    'stock_daily':
        ['stock_daily', 'qt_code', 'table_index', 'stock_basic', '', 'Y', '30'],

    'stock_1min':
        ['stock_1min', 'qt_code', 'table_index', 'stock_basic', '', 'Y', '1'],

    'stock_5min':
        ['stock_5min', 'qt_code', 'table_index', 'stock_basic', '', 'Y', '10'],

    'stock_15min':
        ['stock_15min', 'qt_code', 'table_index', 'stock_basic', '', 'Y', '10'],

    'stock_30min':
        ['stock_30min', 'qt_code', 'table_index', 'stock_basic', '', 'Y', '10'],

    'stock_hourly':
        ['stock_hourly', 'qt_code', 'table_index', 'stock_basic', '', 'Y', '10'],

    'stock_weekly':
        ['stock_weekly', 'qt_code', 'table_index', 'stock_basic', '', 'Y', '100'],

    'stock_monthly':
        ['stock_monthly', 'qt_code', 'table_index', 'stock_basic', '', 'Y', '300'],

    'index_daily':
        ['stock_daily', 'qt_code', 'table_index', 'index_basic', 'SH,SZ', 'Y', '30'],

    'index_1min':
        ['stock_1min', 'qt_code', 'table_index', 'index_basic', 'SH,SZ', 'Y', '1'],

    'index_5min':
        ['stock_5min', 'qt_code', 'table_index', 'index_basic', 'SH,SZ', 'Y', '10'],

    'index_15min':
        ['stock_15min', 'qt_code', 'table_index', 'index_basic', 'SH,SZ', 'Y', '10'],

    'index_30min':
        ['stock_30min', 'qt_code', 'table_index', 'index_basic', 'SH,SZ', 'Y', '10'],

    'index_hourly':
        ['stock_hourly', 'qt_code', 'table_index', 'index_basic', 'SH,SZ', 'Y', '10'],

    'index_weekly':
        ['stock_weekly', 'qt_code', 'table_index', 'index_basic', 'SH,SZ', 'Y', '100'],

    'index_monthly':
        ['stock_monthly', 'qt_code', 'table_index', 'index_basic', 'SH,SZ', 'Y', '300'],

    'fund_daily':
        ['stock_daily', 'qt_code', 'table_index', 'fund_basic', 'SH,SZ', 'Y', '30'],

    'fund_1min':
        ['stock_1min', 'qt_code', 'table_index', 'fund_basic', 'SH,SZ', 'Y', '1'],

    'fund_5min':
        ['stock_5min', 'qt_code', 'table_index', 'fund_basic', 'SH,SZ', 'Y', '10'],

    'fund_15min':
        ['stock_15min', 'qt_code', 'table_index', 'fund_basic', 'SH,SZ', 'Y', '10'],

    'fund_30min':
        ['stock_30min', 'qt_code', 'table_index', 'fund_basic', 'SH,SZ', 'Y', '10'],

    'fund_hourly':
        ['stock_hourly', 'qt_code', 'table_index', 'fund_basic', 'SH,SZ', 'Y', '10'],

    'fund_weekly':
        ['stock_weekly', 'qt_code', 'table_index', 'fund_basic', 'SH,SZ', 'Y', '100'],

    'fund_monthly':
        ['stock_monthly', 'qt_code', 'table_index', 'fund_basic', 'SH,SZ', 'Y', '300'],

}

EASTMONEY_REALTIME_API_MAP = {
    'realtime_bars':  # 实时K线行情数据
        ['real_time_klines', 'qt_code', 'list', 'none', '', 'N', '', ''],

    'realtime_quotes':  # 实时报价数据
        ['real_time_quote', 'qt_code', 'list', 'none', '', 'N', '', '']
}

SINA_API_MAP = {
    'stock_daily':
        ['stock_daily', 'qt_code', 'table_index', 'stock_basic', '', 'Y', '30'],

    'stock_1min':
        ['stock_1min', 'qt_code', 'table_index', 'stock_basic', '', 'Y', '1'],

    'stock_5min':
        ['stock_5min', 'qt_code', 'table_index', 'stock_basic', '', 'Y', '10'],

    'stock_15min':
        ['stock_15min', 'qt_code', 'table_index', 'stock_basic', '', 'Y', '10'],

    'stock_30min':
        ['stock_30min', 'qt_code', 'table_index', 'stock_basic', '', 'Y', '10'],

    'stock_hourly':
        ['stock_hourly', 'qt_code', 'table_index', 'stock_basic', '', 'Y', '10'],

    # SINA does not provide weekly/monthly data
}

SINA_REALTIME_API_MAP = {
    'realtime_bars':  # 实时K线行情数据
        ['realtime_klines', 'qt_code', 'list', 'none', '', 'N', '', ''],

    'realtime_quotes':  # 实时报价数据
        ['realtime_quote', 'qt_code', 'list', 'none', '', 'N', '', '']
}


# coding=utf-8
# ======================================
# File:     test_space.py
# Author:   Jackie PENG
# Contact:  jackie.pengzhao@gmail.com
# Created:  2020-02-12
# Desc:
#   Unittest for all optimizable
#   parameter space related functions.
# ======================================
import unittest

import numpy as np
from numpy import int64
import itertools

from qteasy.space import Space, Parameter, space_around_centre, ResultPool


class TestSpace(unittest.TestCase):
    def test_creation(self):
        """
            test if creation of space object is fine
        """
        # first group of inputs, output Space with two discr axis from [0,10]
        print('testing space objects\n')
        pars_list = [((0, 10), (0, 10)),
                     ([0, 10], [0, 10])]

        types_list = ['int',
                      ['int', 'int']]

        input_pars = itertools.product(pars_list, types_list)
        for p in input_pars:
            print(p)
            s = Space(*p[0], par_types=p[1])
            b = s.boes
            t = s.types
            print(s, t)
            self.assertIsInstance(s, Space)
            self.assertEqual(b, [(0, 10), (0, 10)], 'boes incorrect!')
            self.assertEqual(t, ['int', 'int'], 'types incorrect')

        pars_list = [[(0, 10), (0, 10)],
                     [[0, 10], [0, 10]]]

        types_list = ['foo, bar',
                      ['foo', 'bar']]

        # test invalid par types
        input_pars = itertools.product(pars_list, types_list)
        for p in input_pars:
            self.assertRaises(ValueError, Space, *p[0], par_types=p[1])

        pars_list = [[(0, 10), (0, 10)],
                     [[0, 10], [0, 10]]]

        types_list = [['int', 'enumerate']]

        input_pars = itertools.product(pars_list, types_list)
        for p in input_pars:
            s = Space(*p[0], par_types=p[1])
            b = s.boes
            t = s.types
            self.assertEqual(b, [(0, 10), (0, 10)], 'boes incorrect!')
            self.assertEqual(t, ['int', 'enum'], 'types incorrect')

        pars_list = [(0., 10), (0, 10)]
        s = Space(*pars_list, par_types=None)
        self.assertEqual(s.types, ['float', 'int'])
        self.assertEqual(s.dim, 2)
        self.assertEqual(s.size, (10.0, 11))
        self.assertEqual(s.shape, (np.inf, 11))
        self.assertEqual(s.count, np.inf)
        self.assertEqual(s.boes, [(0., 10), (0, 10)])

        pars_list = [(0., 10), (0, 10)]
        s = Space(*pars_list, par_types=['conti', 'enum'])
        self.assertEqual(s.types, ['float', 'enum'])
        self.assertEqual(s.dim, 2)
        self.assertEqual(s.size, (10.0, 2))
        self.assertEqual(s.shape, (np.inf, 2))
        self.assertEqual(s.count, np.inf)
        self.assertEqual(s.boes, [(0., 10), (0, 10)])

        pars_list = [(1, 2), (2, 3), (3, 4)]
        s = Space(*pars_list)
        self.assertEqual(s.types, ['int', 'int', 'int'])
        self.assertEqual(s.dim, 3)
        self.assertEqual(s.size, (2, 2, 2))
        self.assertEqual(s.shape, (2, 2, 2))
        self.assertEqual(s.count, 8)
        self.assertEqual(s.boes, [(1, 2), (2, 3), (3, 4)])

        pars_list = [(1, 2, 3), (2, 3, 4), (3, 4, 5)]
        s = Space(*pars_list)
        self.assertEqual(s.types, ['enum', 'enum', 'enum'])
        self.assertEqual(s.dim, 3)
        self.assertEqual(s.size, (3, 3, 3))
        self.assertEqual(s.shape, (3, 3, 3))
        self.assertEqual(s.count, 27)
        self.assertEqual(s.boes, [(1, 2, 3), (2, 3, 4), (3, 4, 5)])

        pars_list = [((1, 2, 3), (2, 3, 4), (3, 4, 5))]
        s = Space(*pars_list)
        self.assertEqual(s.types, ['enum'])
        self.assertEqual(s.dim, 1)
        self.assertEqual(s.size, (3,))
        self.assertEqual(s.shape, (3,))
        self.assertEqual(s.count, 3)

        pars_list = ((1, 2, 3), (2, 3, 4), (3, 4, 5))
        s = Space(*pars_list)
        self.assertEqual(s.types, ['enum', 'enum', 'enum'])
        self.assertEqual(s.dim, 3)
        self.assertEqual(s.size, (3, 3, 3))
        self.assertEqual(s.shape, (3, 3, 3))
        self.assertEqual(s.count, 27)
        self.assertEqual(s.boes, [(1, 2, 3), (2, 3, 4), (3, 4, 5)])

        print('testing space_around_centre')
        pars_list = ((0, 10), (0, 10))
        s = Space(*pars_list, par_types=None)
        centre = (5, 5)
        distance = 2
        s2 = space_around_centre(s, centre, distance)
        self.assertEqual(s2.types, ['int', 'int'])
        self.assertEqual(s2.dim, 2)
        self.assertEqual(s2.size, (5, 5))
        self.assertEqual(s2.shape, (5, 5))
        self.assertEqual(s2.count, 25)
        self.assertEqual(s2.boes, [(3, 7), (3, 7)])

        pars_space = ((0., 10.), (0, 10), (0., 10.))
        s = Space(*pars_space, par_types=None)
        centre = (5, 5, 5)
        distance = 2
        s2 = space_around_centre(s, centre, distance)
        self.assertEqual(s2.types, ['float', 'int', 'float'])
        self.assertEqual(s2.dim, 3)
        self.assertEqual(s2.size, (4.0, 5, 4.0))
        self.assertEqual(s2.shape, (np.inf, 5, np.inf))
        self.assertEqual(s2.count, np.inf)
        self.assertEqual(s2.boes, [(3, 7), (3, 7), (3, 7)])

        # test creating Space from Parameter objects
        param1 = Parameter((0, 10), par_type='int')
        param2 = Parameter((0., 10.), par_type='conti')
        param3 = Parameter((1, 2, 3), par_type='enum')
        s = Space(param1, param2, param3)
        self.assertIsInstance(s, Space)
        self.assertEqual(s.types, ['int', 'float', 'enum'])
        self.assertEqual(s.dim, 3)
        self.assertEqual(s.size, (11, 10.0, 3))
        self.assertEqual(s.shape, (11, np.inf, 3))
        self.assertEqual(s.count, np.inf)
        self.assertEqual(s.boes, [(0, 10), (0., 10.), (1, 2, 3)])

    def test_extract(self):
        """

        :return:
        """
        pars_list = [(0, 10), (0, 10)]
        types_list = ['int', 'int']
        s = Space(*pars_list, par_types=types_list)
        extracted_int, count = s.extract(16, 'interval')
        extracted_int_list = list(extracted_int)
        print('extracted int\n', extracted_int_list)
        self.assertEqual(count, 16, 'extraction count wrong!')
        self.assertEqual(extracted_int_list, [(0, 0), (0, 3), (0, 6), (0, 10), (3, 0), (3, 3),
                                              (3, 6), (3, 10), (6, 0), (6, 3), (6, 6), (6, 10),
                                              (10, 0), (10, 3), (10, 6), (10, 10)],
                         'space extraction wrong!')
        extracted_rand, count = s.extract(10, 'rand')
        extracted_rand_list = list(extracted_rand)
        self.assertEqual(count, 10, 'extraction count wrong!')
        print('extracted rand\n', extracted_rand_list)
        for point in list(extracted_rand_list):
            self.assertEqual(len(point), 2)
            self.assertLessEqual(point[0], 10)
            self.assertGreaterEqual(point[0], 0)
            self.assertLessEqual(point[1], 10)
            self.assertGreaterEqual(point[1], 0)

        pars_list = [(0., 10), (0, 10)]
        s = Space(*pars_list, par_types=None)
        extracted_int2, count = s.extract(16, 'interval')
        self.assertEqual(count, 16, 'extraction count wrong!')
        extracted_int_list2 = list(extracted_int2)
        self.assertEqual(extracted_int_list2,
                         [(0, 0), (0, 3), (0, 6), (0, 10),
                          (3.3333333333333335, 0),
                          (3.3333333333333335, 3),
                          (3.3333333333333335, 6),
                          (3.3333333333333335, 10),
                          (6.666666666666667, 0),
                          (6.666666666666667, 3),
                          (6.666666666666667, 6),
                          (6.666666666666667, 10),
                          (10, 0), (10, 3), (10, 6), (10, 10)])
        print('extracted int list 2\n', extracted_int_list2)
        self.assertIsInstance(extracted_int_list2[0][0], float)
        self.assertIsInstance(extracted_int_list2[0][1], (int, int64))
        extracted_rand2, count = s.extract(10, 'rand')
        self.assertEqual(count, 10, 'extraction count wrong!')
        extracted_rand_list2 = list(extracted_rand2)
        print('extracted rand list 2:\n', extracted_rand_list2)
        for point in extracted_rand_list2:
            self.assertEqual(len(point), 2)
            self.assertIsInstance(point[0], float)
            self.assertLessEqual(point[0], 10)
            self.assertGreaterEqual(point[0], 0)
            self.assertIsInstance(point[1], (int, int64))
            self.assertLessEqual(point[1], 10)
            self.assertGreaterEqual(point[1], 0)

        pars_list = [(0., 10), ('a', 'b')]
        s = Space(*pars_list, par_types=['enum', 'enum'])
        extracted_int3, count = s.extract(4, 'interval')
        self.assertEqual(count, 4, 'extraction count wrong!')
        extracted_int_list3 = list(extracted_int3)
        self.assertEqual(extracted_int_list3, [(0., 'a'), (0., 'b'), (10, 'a'), (10, 'b')],
                         'space extraction wrong!')
        print('extracted int list 3\n', extracted_int_list3)
        self.assertIsInstance(extracted_int_list3[0][0], float)
        self.assertIsInstance(extracted_int_list3[0][1], str)
        extracted_rand3, count = s.extract(3, 'rand')
        self.assertEqual(count, 3, 'extraction count wrong!')
        extracted_rand_list3 = list(extracted_rand3)
        print('extracted rand list 3:\n', extracted_rand_list3)
        for point in extracted_rand_list3:
            self.assertEqual(len(point), 2)
            self.assertIsInstance(point[0], (float, int))
            self.assertLessEqual(point[0], 10)
            self.assertGreaterEqual(point[0], 0)
            self.assertIsInstance(point[1], str)
            self.assertIn(point[1], ['a', 'b'])

        pars_list = ((0, 10), (1, 'c'), ('a', 'b'), (1, 14))
        s = Space(pars_list, par_types=['enum'])
        extracted_int4, count = s.extract(1, 'interval')
        self.assertEqual(count, 1, 'extraction count wrong!')
        extracted_int_list4 = list(extracted_int4)
        it = zip(extracted_int_list4, [(0, 10), (1, 'c'), (0, 'b'), (1, 14)])
        for item, item2 in it:
            print(item, item2)
        self.assertTrue(all([tuple(ext_item) == item for ext_item, item in it]))
        print('extracted int list 4\n', extracted_int_list4)
        self.assertIsInstance(extracted_int_list4[0], tuple)
        extracted_rand4, count = s.extract(3, 'rand')
        self.assertEqual(count, 3, 'extraction count wrong!')
        extracted_rand_list4 = list(extracted_rand4)
        print('extracted rand list 4:\n', extracted_rand_list4)
        for point in extracted_rand_list4:
            self.assertEqual(len(point), 2)
            self.assertIsInstance(point[0], (int, str))
            self.assertIn(point[0], [0, 1, 'a'])
            self.assertIsInstance(point[1], (int, str))
            self.assertIn(point[1], [10, 14, 'b', 'c'])
            self.assertIn(point, [(0., 10), (1, 'c'), ('a', 'b'), (1, 14)])

        pars_list = ((0, 10), (1, 'c'), ('a', 'b'), (1, 14)), (1, 4)
        s = Space(*pars_list, par_types=['enum', 'discr'])
        extracted_int5, count = s.extract(16, 'interval')
        self.assertEqual(count, 16, 'extraction count wrong!')
        extracted_int_list5 = list(extracted_int5)
        for item, item2 in extracted_int_list5:
            print(item, item2)
        self.assertTrue(all([tuple(ext_item) == item for ext_item, item in it]))
        print('extracted int list 5\n', extracted_int_list5)
        self.assertIsInstance(extracted_int_list5[0], tuple)
        extracted_rand5, count = s.extract(5, 'rand')
        self.assertEqual(count, 5, 'extraction count wrong!')
        extracted_rand_list5 = list(extracted_rand5)
        print('extracted rand list 5:\n', extracted_rand_list5)
        for point in extracted_rand_list5:
            self.assertEqual(len(point), 2)
            self.assertIsInstance(point[0], tuple)
            print(f'type of point[1] is {type(point[1])}')
            self.assertIsInstance(point[1], (int, np.int64))
            self.assertIn(point[0], [(0., 10), (1, 'c'), ('a', 'b'), (1, 14)])

        print(f'test incremental extraction')
        pars_list = ((10., 250), (10., 250), (10., 250), (10., 250), (10., 250), (10., 250))
        s = Space(*pars_list)
        ext, count = s.extract(4096, 'interval')
        self.assertEqual(count, 4096)
        points = list(ext)
        # 已经取出所有的点，围绕其中10个点生成十个subspaces
        # 检查是否每个subspace都为Space，是否都在s范围内，使用32生成点集，检查生成数量是否正确
        for point in points[1000:1010]:
            subspace = s.from_point(point, 64)
            self.assertIsInstance(subspace, Space)
            self.assertTrue(subspace in s)
            self.assertEqual(subspace.dim, 6)
            self.assertEqual(subspace.types, ['float', 'float', 'float', 'float', 'float', 'float'])
            ext, count = subspace.extract(1024)
            points = list(ext)
            self.assertGreaterEqual(count, 512)
            self.assertLessEqual(count, 4096)
            print(f'\n---------------------------------'
                  f'\nthe space created around point <{point}> is'
                  f'\n{subspace.boes}'
                  f'\nand extracted {count} points, the first 5 are:'
                  f'\n{points[:5]}')

    def test_axis_extract(self):
        # test param object with conti type
        param = Parameter((0., 5))
        self.assertIsInstance(param, Parameter)
        self.assertEqual(param.par_type, 'float')
        self.assertEqual(param.par_range, (0., 5.))
        self.assertEqual(param.count, np.inf)
        self.assertEqual(param.size, 5.0)
        self.assertTrue(np.allclose(param.gen_values(6, 'int'), [0., 1., 2., 3., 4., 5.]))
        self.assertTrue(np.allclose(param.gen_values(11, 'int'), [0., 0.5, 1., 1.5, 2., 2.5, 3., 3.5, 4., 4.5, 5.]))
        extracted = param.gen_values(8, 'rand')
        self.assertEqual(len(extracted), 8)
        self.assertTrue(all([(0 <= item <= 5) for item in extracted]))

        # test param object with discrete type
        param = Parameter((1, 5))
        self.assertIsInstance(param, Parameter)
        self.assertEqual(param.par_type, 'int')
        self.assertEqual(param.par_range, (1, 5))
        self.assertEqual(param.count, 5)
        self.assertEqual(param.size, 5)
        self.assertTrue(np.allclose(param.gen_values(5, 'int'), [1, 2, 3, 4, 5]))
        self.assertRaises(ValueError, param.gen_values, 0.5, 'int')
        extracted = param.gen_values(8, 'rand')
        self.assertEqual(len(extracted), 8)
        self.assertTrue(all([(item in [1, 2, 3, 4, 5]) for item in extracted]))

        # test param object with enumerate type
        param = Parameter((1, 5, 7, 10, 'A', 'F'))
        self.assertIsInstance(param, Parameter)
        self.assertEqual(param.par_type, 'enum')
        self.assertEqual(param.par_range, (1, 5, 7, 10, 'A', 'F'))
        self.assertEqual(param.count, 6)
        self.assertEqual(param.size, 6)
        self.assertEqual(param.gen_values(6, 'int'), [1, 5, 7, 10, 'A', 'F'])
        self.assertRaises(ValueError, param.gen_values, 0.5, 'int')
        extracted = param.gen_values(8, 'rand')
        self.assertEqual(len(extracted), 8)
        self.assertTrue(all([(item in [1, 5, 7, 10, 'A', 'F']) for item in extracted]))

        # test param object with array type
        param = Parameter((0., 1.), par_type='array[2]')
        self.assertIsInstance(param, Parameter)
        self.assertEqual(param.par_type, 'float_array')
        self.assertEqual(param.par_range, (0., 1.))
        self.assertEqual(param.count, np.inf)
        self.assertEqual(param.size, 1.0)
        vals = param.gen_values(5, 'int')
        self.assertEqual(len(vals), 5)
        for val in vals:
            print(f'generated array value: {val}')
            self.assertTrue(isinstance(val, np.ndarray))
            self.assertEqual(val.shape, (2,))
            self.assertTrue(all([(0.0 <= item <= 1.0) for item in val]))


    def test_from_point(self):
        """测试从一个点生成一个space"""
        # 生成一个space，指定space中的一个点以及distance，生成一个sub-space
        pars_list = [(0., 10), (0, 10)]
        s = Space(*pars_list, par_types=None)
        self.assertEqual(s.types, ['float', 'int'])
        self.assertEqual(s.dim, 2)
        self.assertEqual(s.size, (10., 11))
        self.assertEqual(s.shape, (np.inf, 11))
        self.assertEqual(s.count, np.inf)
        self.assertEqual(s.boes, [(0., 10), (0, 10)])

        print('create subspace from a point in space')
        p = (3, 3)
        distance = 2
        subspace = s.from_point(p, distance)
        self.assertIsInstance(subspace, Space)
        self.assertEqual(subspace.types, ['float', 'int'])
        self.assertEqual(subspace.dim, 2)
        self.assertEqual(subspace.size, (4.0, 5))
        self.assertEqual(subspace.shape, (np.inf, 5))
        self.assertEqual(subspace.count, np.inf)
        self.assertEqual(subspace.boes, [(1, 5), (1, 5)])

        print('create subspace from a 6 dimensional discrete space')
        s = Space((10, 250), (10, 250), (10, 250), (10, 250), (10, 250), (10, 250))
        p = (15, 200, 150, 150, 150, 150)
        d = 10
        subspace = s.from_point(p, d)
        self.assertIsInstance(subspace, Space)
        self.assertEqual(subspace.types, ['int', 'int', 'int', 'int', 'int', 'int'])
        self.assertEqual(subspace.dim, 6)
        self.assertEqual(subspace.volume, 65345616)
        self.assertEqual(subspace.size, (16, 21, 21, 21, 21, 21))
        self.assertEqual(subspace.shape, (16, 21, 21, 21, 21, 21))
        self.assertEqual(subspace.count, 65345616)
        self.assertEqual(subspace.boes, [(10, 25), (190, 210), (140, 160), (140, 160), (140, 160), (140, 160)])

        print('create subspace from a 6 dimensional continuous space')
        s = Space((10., 250), (10., 250), (10., 250), (10., 250), (10., 250), (10., 250))
        p = (15, 200, 150, 150, 150, 150)
        d = 10
        subspace = s.from_point(p, d)
        self.assertIsInstance(subspace, Space)
        self.assertEqual(subspace.types, ['float', 'float', 'float', 'float', 'float', 'float'])
        self.assertEqual(subspace.dim, 6)
        self.assertEqual(subspace.volume, 48000000)
        self.assertEqual(subspace.size, (15.0, 20.0, 20.0, 20.0, 20.0, 20.0))
        self.assertEqual(subspace.shape, (np.inf, np.inf, np.inf, np.inf, np.inf, np.inf))
        self.assertEqual(subspace.count, np.inf)
        self.assertEqual(subspace.boes, [(10, 25), (190, 210), (140, 160), (140, 160), (140, 160), (140, 160)])

        print('create subspace with different distances on each dimension')
        s = Space((10., 250), (10., 250), (10., 250), (10., 250), (10., 250), (10., 250))
        p = (15, 200, 150, 150, 150, 150)
        d = [10, 5, 5, 10, 10, 5]
        subspace = s.from_point(p, d)
        self.assertIsInstance(subspace, Space)
        self.assertEqual(subspace.types, ['float', 'float', 'float', 'float', 'float', 'float'])
        self.assertEqual(subspace.dim, 6)
        self.assertEqual(subspace.volume, 6000000)
        self.assertEqual(subspace.size, (15.0, 10.0, 10.0, 20.0, 20.0, 10.0))
        self.assertEqual(subspace.shape, (np.inf, np.inf, np.inf, np.inf, np.inf, np.inf))
        self.assertEqual(subspace.count, np.inf)
        self.assertEqual(subspace.boes, [(10, 25), (195, 205), (145, 155), (140, 160), (140, 160), (145, 155)])

    def test_point_in_space(self):
        sp = Space((0., 10.), (0., 10.), (0., 10.))
        p1 = (5.5, 3.2, 7)
        p2 = (-1, 3, 10)
        self.assertTrue(p1 in sp)
        print(f'point {p1} is in space {sp}')
        self.assertFalse(p2 in sp)
        print(f'point {p2} is not in space {sp}')
        sp = Space((0., 10.), (0., 10.), range(40, 3, -2), par_types=['conti', 'conti', 'enum'])
        p1 = (5.5, 3.2, 8)
        self.assertTrue(p1 in sp)
        print(f'point {p1} is in space {sp}')

        # test array type of point
        sp = Space((0., 10.), (0., 10.), par_types=['array[2]', 'array[2, 2]'])
        p1 = (np.array([5.5, 3.2]), np.array([[1, 2], [3, 4]]))
        self.assertTrue(p1 in sp)
        print(f'point {p1} is in space {sp}')
        p2 = (np.array([5.5, 11]), np.array([[1, 2], [3, 4]]))
        self.assertFalse(p2 in sp)
        print(f'point {p2} is not in space {sp}')

    def test_space_in_space(self):
        print('test if a space is in another space')
        sp = Space((0., 10.), (0., 10.), (0., 10.))
        sp2 = Space((0., 10.), (0., 10.), (0., 10.))
        self.assertTrue(sp2 in sp)
        self.assertTrue(sp in sp2)
        print(f'space {sp2} is in space {sp}\n'
              f'and space {sp} is in space {sp2}\n'
              f'they are equal to each other\n')
        sp2 = Space((0, 5.), (2, 7.), (3., 9.))
        self.assertTrue(sp2 in sp)
        self.assertFalse(sp in sp2)
        print(f'space {sp2} is in space {sp}\n'
              f'and space {sp} is not in space {sp2}\n'
              f'{sp2} is a sub space of {sp}\n')
        sp2 = Space((0, 5), (2, 7), (3., 9))
        self.assertFalse(sp2 in sp)
        self.assertFalse(sp in sp2)
        print(f'space {sp2} is not in space {sp}\n'
              f'and space {sp} is not in space {sp2}\n'
              f'they have different types of axes\n')
        sp = Space((0., 10.), (0., 10.), range(40, 3, -2))
        self.assertFalse(sp in sp2)
        self.assertFalse(sp2 in sp)
        print(f'space {sp2} is not in space {sp}\n'
              f'and space {sp} is not in space {sp2}\n'
              f'they have different types of axes\n')

        # test array type of spaces
        sp = Space((0., 10.), (0., 10.), par_types=['array[2]', 'array[2, 2]'])
        sp2 = Space((0., 5.), (0., 5.), par_types=['array[2]', 'array[2, 2]'])
        self.assertTrue(sp2 in sp)
        print(f'space {sp2} is in space {sp}\n'
              f'space {sp} is a sup space of {sp2}\n')
        # if array dimensions are different, then not in
        sp2 = Space((0., 5.), (0., 5.), par_types=['array[3]', 'array[2, 3, 2]'])
        self.assertFalse(sp2 in sp)
        print(f'space {sp2} is not in space {sp}\n'
              f'space {sp} is not a sup space of {sp2}\n')
        # if array range is larger, then not in
        sp2 = Space((0., 15.), (-5., 5.), par_types=['array[2]', 'array[2, 2]'])
        self.assertFalse(sp2 in sp)
        print(f'space {sp2} is not in space {sp}\n'
              f'space {sp} is not a sup space of {sp2}\n')

    def test_space_around_centre(self):
        sp = Space((0., 10.), (0., 10.), (0., 10.))
        p1 = (5.5, 3.2, 7)
        ssp = space_around_centre(space=sp, centre=p1, radius=1.2)
        print(ssp.boes)
        print('\ntest multiple diameters:')
        self.assertEqual(ssp.boes, [(4.3, 6.7), (2.0, 4.4), (5.8, 8.2)])
        ssp = space_around_centre(space=sp, centre=p1, radius=[1, 2, 1])
        print(ssp.boes)
        self.assertEqual(ssp.boes, [(4.5, 6.5), (1.2000000000000002, 5.2), (6.0, 8.0)])
        print('\ntest points on edge:')
        p2 = (5.5, 3.2, 10)
        ssp = space_around_centre(space=sp, centre=p1, radius=3.9)
        print(ssp.boes)
        self.assertEqual(ssp.boes, [(1.6, 9.4), (0.0, 7.1), (3.1, 10.0)])
        print('\ntest enum spaces')
        sp = Space((0, 100), range(40, 3, -2), par_types=['discr', 'enum'])
        p1 = [34, 12]
        ssp = space_around_centre(space=sp, centre=p1, radius=5, ignore_enums=False)
        self.assertEqual(ssp.boes, [(29, 39), (22, 20, 18, 16, 14, 12, 10, 8, 6, 4)])
        print(ssp.boes)
        print('\ntest enum space and ignore enum axis')
        ssp = space_around_centre(space=sp, centre=p1, radius=5)
        self.assertEqual(ssp.boes, [(29, 39),
                                    (40, 38, 36, 34, 32, 30, 28, 26, 24, 22, 20, 18, 16, 14, 12, 10, 8, 6, 4)])
        print(sp.boes)

    def test_vector_axis_sizes(self):
        """测试 vector_axis_sizes 属性：标量轴为 1，数组轴为 np.prod(shape)。"""
        # int + float + enum -> [1, 1, 1]
        s = Space((1, 10), (3., 10.), (5, 6, 7, 8, 9))
        sizes = s.vector_axis_sizes
        self.assertEqual(sizes, [1, 1, 1], 'vector_axis_sizes for int/float/enum should be [1,1,1]')
        print('vector_axis_sizes int/float/enum:', sizes)

        # 含 array 轴
        s2 = Space((0., 1.), (0., 1.), par_types=['float_array[2]', 'float_array[2, 2]'])
        sizes2 = s2.vector_axis_sizes
        self.assertEqual(sizes2, [2, 4], 'vector_axis_sizes for array[2] and array[2,2] should be [2, 4]')
        print('vector_axis_sizes with array axes:', sizes2)

    def test_point_to_vector_and_vector_to_point(self):
        """测试 point_to_vector / vector_to_point 双向映射及 round-trip。"""
        # int, float, enum
        s = Space((1, 10), (3., 10.), (5, 6, 7, 8, 9))
        p = (5, 6.5, 7)
        self.assertIn(p, s)
        v = s.point_to_vector(p)
        self.assertIsInstance(v, np.ndarray)
        self.assertEqual(v.dtype, np.float64)
        self.assertEqual(v.shape, (3,))
        self.assertEqual(v[0], 5.0)
        self.assertEqual(v[1], 6.5)
        enum_list = list(s.boes[2])
        self.assertEqual(v[2], float(enum_list.index(7)))
        print('point_to_vector result:', v)

        p2 = s.vector_to_point(v)
        self.assertIsInstance(p2, tuple)
        self.assertEqual(len(p2), 3)
        self.assertEqual(p2[0], 5)
        self.assertEqual(p2[1], 6.5)
        self.assertEqual(p2[2], 7)
        self.assertIn(p2, s)
        print('vector_to_point round-trip:', p2)

        # 越界向量解码后应被 clip 到合法范围
        v_oob = np.array([100.0, -1.0, 0.0])
        p3 = s.vector_to_point(v_oob)
        self.assertEqual(p3[0], 10)
        self.assertEqual(p3[1], 3.0)
        self.assertEqual(p3[2], 5)
        self.assertIn(p3, s)
        print('vector_to_point with out-of-bound vector (clipped):', p3)

        # 含 array 轴的空间
        s_arr = Space((0., 1.), (0., 1.), par_types=['float_array[2]', 'float_array[2, 2]'])
        pt_arr = (np.array([0.2, 0.8]), np.array([[0.1, 0.2], [0.3, 0.4]]))
        self.assertIn(pt_arr, s_arr)
        v_arr = s_arr.point_to_vector(pt_arr)
        self.assertEqual(v_arr.size, 2 + 4)
        self.assertAlmostEqual(v_arr[0], 0.2)
        self.assertAlmostEqual(v_arr[1], 0.8)
        pt_arr2 = s_arr.vector_to_point(v_arr)
        self.assertIn(pt_arr2, s_arr)
        np.testing.assert_array_almost_equal(np.asarray(pt_arr2[0]), np.asarray(pt_arr[0]))
        np.testing.assert_array_almost_equal(np.asarray(pt_arr2[1]), np.asarray(pt_arr[1]))
        print('point_to_vector / vector_to_point with array axes: OK')

        # 点不在空间内应报错
        self.assertRaises(AssertionError, s.point_to_vector, (0, 0, 0))  # 0 不在 enum 中
        self.assertRaises(AssertionError, s.point_to_vector, (100, 5, 7))  # 100 超出 int 范围

    def test_neighbors(self):
        """测试 neighbors：仅指定轴变化，返回合法邻域点列表。"""
        s = Space((1, 10), (3., 10.), (5, 6, 7, 8, 9))
        p = (5, 6.5, 7)

        # float 维（axis_index=1）：应得到 ±step 两个候选（若与当前不同）
        nb1 = s.neighbors(p, 1, step=1.0)
        self.assertIsInstance(nb1, list)
        for q in nb1:
            self.assertIn(q, s)
            self.assertEqual(q[0], p[0])
            self.assertEqual(q[2], p[2])
            self.assertNotEqual(q[1], p[1])
        print('neighbors float axis:', nb1)

        # int 维（axis_index=0）
        nb0 = s.neighbors(p, 0)
        self.assertIsInstance(nb0, list)
        for q in nb0:
            self.assertIn(q, s)
            self.assertEqual(q[1], p[1])
            self.assertEqual(q[2], p[2])
        self.assertTrue(any(q[0] == 4 for q in nb0) or any(q[0] == 6 for q in nb0))
        print('neighbors int axis:', nb0)

        # enum 维（axis_index=2）
        nb2 = s.neighbors(p, 2, count=2)
        self.assertIsInstance(nb2, list)
        for q in nb2:
            self.assertIn(q, s)
            self.assertEqual(q[0], p[0])
            self.assertEqual(q[1], p[1])
            self.assertNotEqual(q[2], p[2])
            self.assertIn(q[2], (5, 6, 8, 9))
        print('neighbors enum axis:', nb2)

        # 边界点：int 在 1 时只有 +1 邻域
        p_edge = (1, 5.0, 5)
        nb_edge = s.neighbors(p_edge, 0)
        self.assertIn((2, 5.0, 5), nb_edge)
        print('neighbors at int boundary:', nb_edge)

        # float_array 维（axis_index=1）：邻域为同形状数组且仅该维与 point 不同
        s_farr = Space((0, 5), (0., 1.), par_types=['int', 'float_array[2]'])
        p_farr = (2, np.array([0.3, 0.7]))
        self.assertIn(p_farr, s_farr)
        nb_farr = s_farr.neighbors(p_farr, 1, count=3)
        self.assertIsInstance(nb_farr, list)
        for q in nb_farr:
            self.assertIn(q, s_farr)
            self.assertEqual(q[0], p_farr[0])
            self.assertEqual(np.asarray(q[1]).shape, (2,))
            self.assertFalse(np.array_equal(np.asarray(q[1]), np.asarray(p_farr[1])))
        print('neighbors float_array axis:', [tuple(np.asarray(x[1]) for x in nb_farr)])

        # int_array 维（axis_index=1）：同上，仅该维为整型数组且与 point 不同
        s_iarr = Space((0, 5), (0, 10), par_types=['int', 'int_array[2]'])
        p_iarr = (2, np.array([3, 7]))
        self.assertIn(p_iarr, s_iarr)
        nb_iarr = s_iarr.neighbors(p_iarr, 1, count=3)
        self.assertIsInstance(nb_iarr, list)
        for q in nb_iarr:
            self.assertIn(q, s_iarr)
            self.assertEqual(q[0], p_iarr[0])
            self.assertEqual(np.asarray(q[1]).shape, (2,))
            self.assertFalse(np.array_equal(np.asarray(q[1]), np.asarray(p_iarr[1])))
        print('neighbors int_array axis:', [tuple(np.asarray(x[1]) for x in nb_iarr)])

    def test_sample_subspace(self):
        """测试 sample_subspace：在中心邻域内采样 count 个点。"""
        s = Space((0, 10), (0., 10.))
        center = (5, 5.0)
        pts = s.sample_subspace(center, radius=2, count=5)
        self.assertIsInstance(pts, list)
        self.assertEqual(len(pts), 5)
        for q in pts:
            self.assertIn(q, s)
            self.assertGreaterEqual(q[0], 3)
            self.assertLessEqual(q[0], 7)
            self.assertGreaterEqual(q[1], 3.0)
            self.assertLessEqual(q[1], 7.0)
        print('sample_subspace int/float:', pts)

        # 含 enum：ignore_enums=True 时 enum 维保持全枚举
        s2 = Space((0, 5), (0., 5.), ('a', 'b', 'c'))
        center2 = (2, 2.5, 'b')
        pts2 = s2.sample_subspace(center2, radius=1, count=4, ignore_enums=True)
        self.assertEqual(len(pts2), 4)
        for q in pts2:
            self.assertIn(q, s2)
            self.assertIn(q[2], ('a', 'b', 'c'))
        print('sample_subspace with enum (ignore_enums=True):', pts2)

        # 中心不在空间内应报错
        self.assertRaises(AssertionError, s.sample_subspace, (100, 5.0), 1, 3)

        # float_array：子空间为每元素在 center±radius 内裁剪到轴边界
        s_fa = Space((0., 1.), (0., 1.), par_types=['float_array[2]', 'float_array[2, 2]'])
        center_fa = (np.array([0.3, 0.7]), np.array([[0.2, 0.3], [0.4, 0.5]]))
        self.assertIn(center_fa, s_fa)
        pts_fa = s_fa.sample_subspace(center_fa, radius=0.2, count=5)
        self.assertEqual(len(pts_fa), 5)
        for q in pts_fa:
            self.assertIn(q, s_fa)
            a1, a2 = np.asarray(q[0]), np.asarray(q[1])
            self.assertEqual(a1.shape, (2,))
            self.assertEqual(a2.shape, (2, 2))
            # 元素应在 [center - 0.2, center + 0.2] 裁剪到 [0, 1] 内
            self.assertTrue(np.all(a1 >= 0) and np.all(a1 <= 1))
            self.assertTrue(np.all(a2 >= 0) and np.all(a2 <= 1))
        print('sample_subspace float_array:', tuple(np.asarray(x[0]) for x in pts_fa))

        # int_array：标量 radius，子空间为每元素在 center±radius 内裁剪
        s_ia = Space((0, 10), (0, 10), par_types=['int_array[2]', 'int_array[2]'])
        center_ia = (np.array([4, 6]), np.array([2, 8]))
        self.assertIn(center_ia, s_ia)
        pts_ia = s_ia.sample_subspace(center_ia, radius=2, count=4)
        self.assertEqual(len(pts_ia), 4)
        for q in pts_ia:
            self.assertIn(q, s_ia)
            a1, a2 = np.asarray(q[0]), np.asarray(q[1])
            self.assertEqual(a1.shape, (2,))
            self.assertEqual(a2.shape, (2,))
            self.assertTrue(np.all(a1 >= 2) and np.all(a1 <= 8))
            self.assertTrue(np.all(a2 >= 0) and np.all(a2 <= 10))
        print('sample_subspace int_array:', tuple(np.asarray(x[0]) for x in pts_ia))

        # 按维 radius（list）：对 array 维使用对应半径
        radius_per_dim = [0.15, 0.25]  # 两维均为 array
        pts_fa2 = s_fa.sample_subspace(center_fa, radius=radius_per_dim, count=3)
        self.assertEqual(len(pts_fa2), 3)
        for q in pts_fa2:
            self.assertIn(q, s_fa)
        print('sample_subspace float_array with list radius:', len(pts_fa2))


class TestPool(unittest.TestCase):
    def setUp(self):
        self.p = ResultPool(5)
        self.items = ['first', 'second', (1, 2, 3), 'this', 24]
        self.perfs = [1, 2, 3, 4, 5]
        self.additional_result1 = ('abc', 12)
        self.additional_result2 = ([1, 2], -1)
        self.additional_result3 = (12, 5)

    def test_create(self):
        self.assertIsInstance(self.p, ResultPool)

    def test_operation(self):
        self.p.push(self.additional_result1[0], self.additional_result1[1])
        self.p.cut()
        self.assertEqual(self.p.item_count, 1)
        self.assertEqual(self.p.items, ['abc'])
        for item, perf in zip(self.items, self.perfs):
            self.p.push(item, perf)
        self.assertEqual(self.p.item_count, 6)
        self.assertEqual(self.p.items, ['abc', 'first', 'second', (1, 2, 3), 'this', 24])
        self.p.cut()
        self.assertEqual(self.p.items, ['second', (1, 2, 3), 'this', 24, 'abc'])
        self.assertEqual(self.p.perfs, [2, 3, 4, 5, 12])

        self.p.push(self.additional_result2[0], self.additional_result2[1])
        self.p.push(self.additional_result3[0], self.additional_result3[1])
        self.assertEqual(self.p.item_count, 7)
        self.p.cut(keep_largest=False)
        self.assertEqual(self.p.items, [[1, 2], 'second', (1, 2, 3), 'this', 24])
        self.assertEqual(self.p.perfs, [-1, 2, 3, 4, 5])


if __name__ == '__main__':
    unittest.main()
    


# coding=utf-8
# ======================================
# File:     test_user_defined.py
# Author:   Jackie PENG
# Contact:  jackie.pengzhao@gmail.com
# Created:  2020-02-12
# Desc:
#   Test file for user-defined
# strategies.
# ======================================
import unittest

from numba.core.utils import benchmark

import qteasy as qt
import numpy as np

from qteasy.parameter import Parameter
from qteasy.datatypes import StgData


def market_value_weighted(stock_return, mv, mv_cat, bp_cat, mv_target, bp_target):
    """ 根据mv_target和bp_target计算市值加权收益率

    """
    sel = (mv_cat == mv_target) & (bp_cat == bp_target)
    mv_total = np.nansum(mv[sel])
    mv_weight = mv / mv_total
    return_total = np.nansum(stock_return[sel] * mv_weight[sel])
    return return_total


class MultiFactors(qt.FactorSorter):

    def __init__(self, par_values: tuple = (0.5, 0.3, 0.7)):
        super().__init__(
                pars=[Parameter((0.01, 0.99), name='size_gate_percentile', par_type='float'),
                      Parameter((0.01, 0.49), name='bp_small_percentile', par_type='float'),
                      Parameter((0.50, 0.99), name='bp_large_percentile', par_type='float')],
                name='MultiFactor',
                description='根据Fama-French三因子回归模型估算HS300成分股的alpha值选股',
                data_types=[StgData('pb', freq='d', asset_type='E'),
                            StgData('total_mv', freq='d', asset_type='E'),
                            StgData('close', freq='d', asset_type='E'),
                            StgData('close-000300.SH', freq='d', asset_type='IDX')],  # 参考数据
                window_length=20,
                use_latest_data_cycle=True,
                max_sel_count=10,  # 最多选出10支股票
                sort_ascending=True,  # 选择因子最小的股票
                condition='less',  # 仅选择因子小于某个值的股票
                lbound=0,  # 仅选择因子小于0的股票
                ubound=0,  # 仅选择因子小于0的股票
        )
        if par_values:
            self.update_par_values(*par_values)

    def realize(self):

        size_gate_percentile, bp_small_percentile, bp_large_percentile = self.get_pars('size_gate_percentile',
                                                                                       'bp_small_percentile',
                                                                                       'bp_large_percentile')
        pb_data, mv_data, close_data = self.get_data('pb_E_d', 'total_mv_E_d', 'close_E_d')  # 获取策略数据

        # 读取投资组合的数据PB和total_MV的最新值
        pb = pb_data[-1]  # 当前所有股票的PB值
        mv = mv_data[-1]  # 当前所有股票的市值
        pre_close = close_data[-2]  # 当前所有股票的前收盘价
        close = close_data[-1]  # 当前所有股票的最新收盘价

        # 读取参考数据(r)
        r = self.get_data('close-000300.SH_IDX_d')  # 获取参考数据
        market_pre_close = r[-2]  # HS300的昨收价
        market_close = r[-1]  # HS300的收盘价

        # 计算账面市值比，为pb的倒数
        bp = pb ** -1
        # 计算市值的50%的分位点,用于后面的分类
        size_gate = np.nanquantile(mv, size_gate_percentile)
        # 计算账面市值比的30%和70%分位点,用于后面的分类
        bm_30_gate = np.nanquantile(bp, bp_small_percentile)
        bm_70_gate = np.nanquantile(bp, bp_large_percentile)
        # 计算每只股票的当日收益率
        stock_return = pre_close / close - 1

        # 根据每只股票的账面市值比和市值，给它们分配bp分类和mv分类
        # 市值小于size_gate的cat为1，否则为2
        mv_cat = np.ones_like(mv)
        mv_cat += (mv > size_gate).astype('float')
        # bp小于30%的cat为1，30%～70%之间为2，大于70%为3
        bp_cat = np.ones_like(bp)
        bp_cat += (bp > bm_30_gate).astype('float')
        bp_cat += (bp > bm_70_gate).astype('float')

        # 获取小市值组合的市值加权组合收益率
        smb_s = (market_value_weighted(stock_return, mv, mv_cat, bp_cat, 1, 1) +
                 market_value_weighted(stock_return, mv, mv_cat, bp_cat, 1, 2) +
                 market_value_weighted(stock_return, mv, mv_cat, bp_cat, 1, 3)) / 3
        # 获取大市值组合的市值加权组合收益率
        smb_b = (market_value_weighted(stock_return, mv, mv_cat, bp_cat, 2, 1) +
                 market_value_weighted(stock_return, mv, mv_cat, bp_cat, 2, 2) +
                 market_value_weighted(stock_return, mv, mv_cat, bp_cat, 2, 3)) / 3
        smb = smb_s - smb_b
        # 获取大账面市值比组合的市值加权组合收益率
        hml_b = (market_value_weighted(stock_return, mv, mv_cat, bp_cat, 1, 3) +
                 market_value_weighted(stock_return, mv, mv_cat, bp_cat, 2, 3)) / 2
        # 获取小账面市值比组合的市值加权组合收益率
        hml_s = (market_value_weighted(stock_return, mv, mv_cat, bp_cat, 1, 1) +
                 market_value_weighted(stock_return, mv, mv_cat, bp_cat, 2, 1)) / 2
        hml = hml_b - hml_s

        # 计算市场收益率
        market_return = market_pre_close / market_close - 1

        coff_pool = []
        # 对每只股票进行回归获取其alpha值
        for rtn in stock_return:
            x = np.array([[market_return, smb, hml, 1.0]])
            y = np.array([[rtn]])
            # OLS估计系数
            coff = np.linalg.lstsq(x, y)[0][3][0]
            coff_pool.append(coff)

        # 以alpha值为股票组合的选股因子执行选股
        factors = np.array(coff_pool)

        return factors


class IndexEnhancement(qt.GeneralStg):

    def __init__(self, par_values: tuple = (0.35, 0.8, 5)):
        super().__init__(
                # 参数1:沪深300指数权重阈值，低于它的股票不被选中，参数2: 初始权重，参数3: 连续涨跌天数，作为强弱势判断阈值
                pars=[Parameter((0.01, 0.99), name='weight_threshold', par_type='float'),
                      Parameter((0.51, 0.99), name='initial_weight', par_type='float'),
                      Parameter((2, 20), name='price_days', par_type='int')],
                name='IndexEnhancement',
                description='跟踪HS300指数选股，并根据连续上涨/下跌趋势判断强弱势以增强权重',
                # 利用HS300权重设定选股权重, 根据收盘价判断强弱势
                data_types=[StgData('wt_idx|000300.SH', freq='m', asset_type='ANY', window_length=1),
                            StgData('close', freq='d', asset_type='ANY', window_length=20, )],
                use_latest_data_cycle=True,
        )
        if par_values:
            self.update_par_values(*par_values)

    def realize(self):
        weight_threshold, init_weight, price_days = self.get_pars('weight_threshold', 'initial_weight', 'price_days')
        index_weight, prices = self.get_data('wt_idx|000300.SH_ANY_m', 'close_ANY_d')
        # 读取投资组合的权重wt和最近price_days天的收盘价
        wt = index_weight[-1]  # 当前所有股票的权重值
        pre_close = prices[-price_days - 1:-1]
        close = prices[-price_days:]  # 当前所有股票的最新连续收盘价
        # 计算连续price_days天的收益
        stock_returns = pre_close - close  # 连续p天的收益

        # 设置初始选股权重为0.8
        weights = init_weight * np.ones_like(wt)

        # 剔除所有没有权重（NaN）的股票
        weights[np.where(np.isnan(wt))] = 0

        # 剔除掉权重小于weight_threshold的股票
        weights[wt < weight_threshold] = 0

        # 找出强势股，将其权重设为1, 找出弱势股，将其权重设置为 init_weight - (1 - init_weight)
        up_trends = np.all(stock_returns > 0, axis=0)
        weights[up_trends] = 1.0
        down_trend_weight = init_weight - (1 - init_weight)
        down_trends = np.all(stock_returns < 0, axis=0)
        weights[down_trends] = down_trend_weight

        # 实际选股权重为weights * HS300权重
        weights *= wt
        # print(f'select weight is: \n{weights}')

        return weights


class GridTrading(qt.GeneralStg):

    def __init__(self, par_values: tuple = (2.0, 3.0, 0.3, 0.5, 300)):
        super().__init__(
                # 仓位配置的阈值：参数1:低仓位阈值，参数2: 高仓位阈值，参数3：低仓位比例，参数4:高仓位比例，参数5:计算天数
                pars=[Parameter((0.5, 3.0), name='low_threshold', par_type='float'),
                      Parameter((2.0, 10.0), name='high_threshold', par_type='float'),
                      Parameter((0.01, 0.5), name='low_position', par_type='float'),
                      Parameter((0.5, 0.99), name='high_position', par_type='float'),
                      Parameter((10, 300), name='days', par_type='int')],
                name='GridTrading',
                description='根据过去300份钟的股价均值和标准差，改变投资金额的仓位',
                data_types=[StgData('close', freq='1min', asset_type='ANY')],  # 使用分钟收盘价调整
                window_length=300,
                use_latest_data_cycle=False,  # 高频数据不需要使用当前数据区间
        )
        if par_values:
            self.update_par_values(*par_values)

    def realize(self):
        """策略输出PT信号，即仓位目标信号"""

        low_threshold, high_threshold, low_pos, hi_pos, days = self.get_pars('low_threshold',
                                                                             'high_threshold',
                                                                             'low_position',
                                                                             'high_position',
                                                                             'days')
        h = self.get_data('close_ANY_1min')

        # 读取最近N天的收盘价
        close = h[- days:]  # 最新连续收盘价
        current_close = h[-1]  # 当天的收盘价

        # 计算N天的平均价和标准差，并计算仓位阈值
        close_mean = np.nanmean(close)
        close_std = np.nanstd(close)
        hi_positive = close_mean + high_threshold * close_std
        low_positive = close_mean + low_threshold * close_std
        low_negative = close_mean - low_threshold * close_std
        hi_negative = close_mean - high_threshold * close_std

        # 根据当前的实际价格确定目标仓位，并将目标仓位作为信号输出
        pos = np.zeros_like(close_mean)
        pos = np.where(current_close > hi_positive, hi_pos, pos)
        pos = np.where((hi_positive >= current_close > low_positive), low_pos, pos)
        pos = np.where((low_positive >= current_close > low_negative), 0, pos)
        pos = np.where((low_negative >= current_close > hi_negative), - low_pos, pos)
        pos = np.where(current_close >= hi_negative, - hi_pos, pos)

        return pos


class AlphaPT(qt.GeneralStg):

    def realize(self):
        # 从历史数据编码中读取四种历史数据的最新数值
        total_mv = self.get_data('total_mv_E_d')[-1]  # 总市值
        total_liab = self.get_data('total_liab_E_q')[-1]  # 总负债
        cash_equ = self.get_data('c_cash_equ_end_period_E_q')[-1]  # 现金及现金等价物总额
        ebitda = self.get_data('ebitda_E_q')[-1]  # ebitda，息税折旧摊销前利润

        # 选股因子为EV/EBIDTA，使用下面公式计算
        factors = (total_mv + total_liab - cash_equ) / ebitda
        # 处理交易信号，将所有小于0的因子变为NaN
        factors = np.where(factors < 0, np.nan, factors)
        # 选出数值最小的30个股票的序号
        arg_partitioned = factors.argpartition(30)
        selected = arg_partitioned[:30]  # 被选中的30个股票的序号，此时股票可能有NaN被选中的情况，需要去掉
        not_selected = arg_partitioned[30:]  # 未被选中的其他股票的序号（包括因子为NaN的股票）

        # 如果选出的股票中有因子为NaN的，则剔除掉
        selected = selected[~np.isnan(selected)]
        sel_count = len(selected)

        # 开始生成PT交易信号
        signal = np.zeros_like(factors)
        # 所有被选中的股票的持仓目标被设置为0.03，表示持有3.3%
        signal[selected] = 1 / sel_count
        # 其余未选中的所有股票持仓目标在PT信号模式下被设置为0，代表目标仓位为0
        signal[not_selected] = 0

        return signal


class FastExperiments(unittest.TestCase):
    """This test case is created to have experiments done that can be quickly called from Command line"""

    def setUp(self):
        from qteasy import DataSource, QT_CONFIG
        self.test_ds = DataSource(
                source_type='db',
                host=QT_CONFIG['test_db_host'],
                port=QT_CONFIG['test_db_port'],
                user=QT_CONFIG['test_db_user'],
                password=QT_CONFIG['test_db_password'],
                db_name=QT_CONFIG['test_db_name']
        )

    def test_multi_factors(self):
        """ test strategy MultiFactors"""
        self.assertEqual(1, 1)
        shares = qt.filter_stock_codes(index='000300.SH', date='20210101')
        print(len(shares), shares[:10])

        alpha = MultiFactors()
        op = qt.Operator(alpha,
                         signal_type='PT',
                         run_timing='close',  # 在周期结束（收盘）时运行
                         run_freq='M',  # 每月执一次选股（每周或每天都可以）
                         )

        op.set_blender("0.8*s0", 'Group_1')
        qt.run(op=op,
               mode=1,
               invest_start='20210101',
               invest_end='20220501',
               asset_type='E',
               invest_cash_amounts=[1000000],
               asset_pool=shares,
               trade_batch_size=100,
               sell_batch_size=1,
               trade_log=True)
        self.assertTrue(True)

    def test_index_enhancement(self):
        """ tests for self-defined strategies"""
        self.assertEqual(1, 1)

        shares = qt.filter_stock_codes(index='000300.SH', date='20210101')
        print(len(shares), shares[:10])

        alpha = IndexEnhancement()
        op = qt.Operator(alpha,
                         signal_type='PT',
                         run_timing='close',  # 在周期结束（收盘）时运行
                         run_freq='d',  # 每天执行一次选股
                         )

        op.op_type = 'stepwise'
        op.set_blender("0.8*s0", 'Group_1')
        qt.run(op=op,
               mode=1,
               invest_start='20210101',
               invest_end='20220501',
               asset_type='E',
               invest_cash_amounts=[1000000],
               asset_pool=shares,
               trade_batch_size=100,
               sell_batch_size=1,
               trade_log=True)
        self.assertTrue(True)

    def test_grid_trading(self):
        """ test for strategy GridTrading"""
        self.assertEqual(1, 1)

        alpha = GridTrading()
        op = qt.Operator(alpha, signal_type='PT', run_timing='close', run_freq='1min')

        op.set_blender("1.0*s0", 'Group_1')
        qt.run(
                op=op,
                mode=1,
                invest_start='20220401',
                invest_end='20220731',
                invest_cash_amounts=[1000000],
                asset_type='IDX',
                asset_pool=['000300.SH'],
                trade_batch_size=1.,
                sell_batch_size=1.,
                trade_log=True,
                allow_sell_short=True,
        )
        self.assertTrue(True)

    def test_get_history_data(self):
        """
        """
        # gen_value normal data
        data = qt.get_history_data(htypes='open, high, low, close',
                                   start='20211230',
                                   end='20220110',
                                   shares='000001.SZ',
                                   freq='d',
                                   adj='n',
                                   asset_type='E')
        print(f'not adjusted data: \n{data}')
        # gen_value back adjusted price for one asset
        data = qt.get_history_data(htypes='open, high, low, close',
                                   start='20211230',
                                   end='20220110',
                                   shares='000001.SZ',
                                   freq='d',
                                   adj='b',
                                   asset_type='E')
        print(f'adjusted data: \n{data}')
        # gen_value back adjusted prices for both assets mixed
        data = qt.get_history_data(htypes='open, high, low, close',
                                   start='20211230',
                                   end='20220110',
                                   shares='000001.SZ, 512000.SH',
                                   freq='d',
                                   adj='b',
                                   asset_type='E, FD')
        print(f'adjusted data: \n{data}')
        self.assertTrue(True)

    def test_fast_experiments(self):
        """ test for strategy GridTrading"""
        shares = qt.filter_stock_codes(index='000300.SH', date='20190501')  # 选择股票池，包括2019年5月以来所有沪深300指数成分股
        # 设置回测的运行参数
        # shares = ['000001.SZ', '000002.SZ', '000063.SZ', '000069.SZ', '000100.SZ', '000157.SZ', '000166.SZ',
        #           '000333.SZ', '000338.SZ', '000413.SZ', '000415.SZ', '000423.SZ', '000425.SZ', '000538.SZ',
        #           '000568.SZ', '000596.SZ', '000625.SZ', '000627.SZ', '000629.SZ', '000630.SZ', '000651.SZ',
        #           '000656.SZ', '000661.SZ', '000671.SZ', '000703.SZ', '000709.SZ', '000723.SZ', '000725.SZ',
        #           '000728.SZ', '000768.SZ', '000776.SZ', '000783.SZ', '000786.SZ', '000858.SZ', '000876.SZ',
        #           '000895.SZ', '000898.SZ', '000938.SZ', '000961.SZ', '000963.SZ', '001979.SZ', '002001.SZ',
        #           '002007.SZ', '002008.SZ', '002010.SZ', '002024.SZ', '002027.SZ', '002032.SZ', '002044.SZ',
        #           '002050.SZ']
        qt.configure(mode=1,  # mode=1表示回测模式
                     invest_start='20160405',  # 回测开始日期
                     invest_end='20210201',  # 回测结束日期
                     asset_type='E',  # 投资品种为股票
                     backtest_price_adj='b',
                     asset_pool=shares,  # shares包含同期沪深300指数的成份股
                     trade_batch_size=100,  # 买入批量为100股
                     sell_batch_size=1,  # 卖出批量为整数股
                     trade_log=True,  # 生成交易记录
                     )

        # alpha = MultiFactors()  # 生成一个交易策略的实例，名为alpha
        op = qt.Operator('dma', signal_type='PT',
                         run_timing='close',  # 在周期结束（收盘）时运行
                         run_freq='M',  # 每月执行一次选股（每周或每天都可以）
                         )  # 生成交易员对象，操作alpha策略，交易信号的类型为‘PT'，意思是生成的信号代表持仓比例，例如1代表100%持有股票，0.35表示持有股票占资产的35%
        # op.set_blender('1.0*s0')  # 交易策略混合方式，只有一个策略，不需要混合
        qt.run(op=op)  # 开始运行


if __name__ == '__main__':
    unittest.main()